# -*- coding: utf-8 -*-

import locale
import collections

from datetime import datetime
from typing import Optional

from repository import OrderRepository
from helper import remove_accents


DOUBLE_LINE = u"======================================\n"


class DiscountReport(object):
    TYPE_REAL_DATE = "RealDate"
    TYPE_BUSINESS_PERIOD = "BusinessPeriod"
    TYPE_SESSION_ID = "SessionId"

    def __init__(self, order_repository, store_id):
        # type: (OrderRepository, unicode) -> None
        self.order_repository = order_repository
        self.store_id = store_id

        try:
            locale.setlocale(locale.LC_ALL, "portuguese_brazil")
        except Exception as _:
            locale.setlocale(locale.LC_ALL, "pt_BR.UTF-8")

    def generate_discount_report_by_real_date(self, pos_id, report_pos, initial_date, end_date, operator_id):
        # type: (int, Optional[unicode], datetime, datetime, Optional[unicode]) -> str
        return self._generate_discount_report(self.TYPE_REAL_DATE, pos_id, report_pos, initial_date, end_date,
                                              operator_id, None)

    def generate_discount_report_by_business_period(self, pos_id, report_pos, initial_date, end_date, operator_id):
        # type: (int, Optional[unicode], datetime, datetime, unicode) -> str

        return self._generate_discount_report(self.TYPE_BUSINESS_PERIOD, pos_id, report_pos, initial_date, end_date,
                                              operator_id, None)

    def generate_discount_report_by_session_id(self, pos_id, session_id):
        # type: (int, Optional[unicode]) -> str

        return self._generate_discount_report(self.TYPE_SESSION_ID, pos_id, None, None, None, None, session_id)

    def _generate_discount_report(self, report_type, pos_id, report_pos, initial_date, end_date, operator_id,
                                  session_id):
        # type: (unicode, int, Optional[unicode], Optional[datetime], Optional[datetime], Optional[unicode], Optional[unicode]) -> str # noqa

        discount_orders = None
        if report_type == self.TYPE_REAL_DATE:
            discount_orders = self.order_repository.get_discount_orders_by_real_date(initial_date, end_date,
                                                                                     operator_id, report_pos)

        elif report_type == self.TYPE_BUSINESS_PERIOD:
            discount_orders = self.order_repository.get_discount_orders_by_business_period(initial_date, end_date,
                                                                                           operator_id, report_pos)

        elif report_type == self.TYPE_SESSION_ID:
            discount_orders = self.order_repository.get_discount_orders_by_session_id(session_id)
            start_index = session_id.find("user=")
            end_index = session_id.find(",", start_index)
            operator_id = session_id[start_index + 5:end_index]
            start_index = session_id.find("period=")
            initial_date = end_date = datetime.strptime(session_id[start_index + 7:], "%Y%m%d")

        header = self._build_header(report_type, pos_id, report_pos, initial_date, end_date, operator_id)
        body = self._build_body(discount_orders)

        report = header + body
        report_bytes = report.encode("utf-8")

        return report_bytes

    def _build_header(self, report_type, pos_id, report_pos, initial_date, end_date, operator_id):
        operator = "Todos"
        if operator_id is not None:
            operator = operator_id

        pos_list = "Todos" if report_pos is None else report_pos.zfill(2)
        if report_type == self.TYPE_REAL_DATE:
            report_type_text = "Data Fiscal"
        elif report_type == self.TYPE_BUSINESS_PERIOD:
            report_type_text = "Dia Negocio"
        else:
            report_type_text = "Surpresa"
            pos_list = pos_id

        header = DOUBLE_LINE
        header += u"  Relatorio Descontos - {0:>11}\n".format(report_type_text)
        header += u" Loja.........: " + self.store_id + "\n"
        header += u" Data/hora....: " + datetime.now().strftime("%d/%m/%Y %H:%M:%S") + "\n"
        formatted_initial_date = initial_date.strftime("%d/%m/%y")
        formatted_end_date = end_date.strftime("%d/%m/%y")
        header += u" {0:.<13}: {1} - {2}\n".format(report_type_text, formatted_initial_date, formatted_end_date)
        header += u" ID Operador..: " + u"{0} (POS # {1})\n".format(operator, str(pos_id).zfill(2))
        temp_pos_list = u" POS incluidos: " + u"{0}".format(pos_list)
        header += temp_pos_list[:38]
        if len(temp_pos_list) > 38:
            header += (u"\n                %s" % temp_pos_list[38:60])
        if len(temp_pos_list) > 66:
            header += (u"\n                %s" % temp_pos_list[60:82])
        header += u"\n"

        return header

    @staticmethod
    def _build_body(orders_with_discount):
        manual_orders_with_discount = [x for x in orders_with_discount if not x.delivery]
        delivery_orders_with_discount = [x for x in orders_with_discount if x.delivery]

        body = DOUBLE_LINE
        body += u"    Descricao       Qtd      Total\n"
        body += DiscountReport._add_discount_total_line("Total Descontos..:", orders_with_discount)
        body += DiscountReport._add_discount_total_line("      Delivery...:", delivery_orders_with_discount)
        body += DiscountReport._add_discount_total_line("      Operador...:", manual_orders_with_discount)
        body += DOUBLE_LINE
        body += u"PEDIDOS".center(38)
        body += DOUBLE_LINE

        discounts_by_date = DiscountReport._group_discounts_by_date(manual_orders_with_discount)

        for key in discounts_by_date.keys():
            body += "\n\n" + _format_date_only(key).center(38)
            total_discounts = len(discounts_by_date[key])
            discount_count = 1
            for order in discounts_by_date[key]:
                write_end_line = total_discounts != discount_count
                body = DiscountReport._add_discount(body, order, write_end_line)
                discount_count += 1

            body += DOUBLE_LINE

        return body

    @staticmethod
    def _add_discount_total_line(text, orders):
        discount_qty = len(orders)
        total_discount_value = sum([order.discount_amount for order in orders])

        return u"{0} [{1:>4}] R${2:>10}\n".format(text, discount_qty, _float_convert(total_discount_value))

    @staticmethod
    def _add_discount(body, order, write_end_line):
        order_id = str(order.order_id)
        applied_date = _format_time(order.applied_date)
        total_gross = _float_convert(order.total_gross)
        discount_amount = _float_convert(order.discount_amount)
        operator = remove_accents(order.operator[:20])
        authorizer = remove_accents(order.authorizer[:20])

        body += "\nPed./Hora..: {0:>12} /   {1:>8}\n".format(order_id, applied_date)
        body += "Total/Desc.:   R${0:>8} / R${1:>8}\n".format(total_gross, discount_amount)
        body += "Operator...: {0:>25}\n".format(operator)
        body += "Autorizador: {0:>25}\n".format(authorizer)

        if write_end_line:
            body += u"--------------------------------------\n"

        return body

    @staticmethod
    def _group_discounts_by_date(discount_orders):
        discounts_by_date = {}
        for discount in discount_orders:
            date = datetime(discount.applied_date.year, discount.applied_date.month, discount.applied_date.day)
            if date in discounts_by_date:
                discounts_by_date[date].append(discount)
            else:
                discounts_by_date[date] = [discount]
        return collections.OrderedDict(sorted(discounts_by_date.items()))

    @staticmethod
    def _get_discount_values(discount_orders):
        value_discount = 0
        qty_discount = 0

        for order in discount_orders:
            if order.discount > 0:
                value_discount += order.discount
                qty_discount += 1

        return qty_discount, value_discount


def _format_date_only(date):
    return datetime.strftime(date, "%d/%m/%Y")


def _format_time(date):
    return datetime.strftime(date, "%H:%M:%S")


def _float_convert(value):
    return locale.format("%.2f", value, True)



from datetime import datetime, timedelta

import iso8601
from application.model import \
    BenefitControllerDto, \
    OperationStatus, \
    BenefitAppliers, \
    OperationDescription
from application.model.configuration import Configurations
from mbcontextmessagehandler import MbContextMessageBus
from mwhelper import BaseRepository
from persistence import Connection
from typing import List, Optional


class DBRepository(BaseRepository):

    def __init__(self, configs, message_bus):
        # type: (Configurations, MbContextMessageBus) -> None

        super(DBRepository, self).__init__(message_bus.mbcontext)

        self.logger = configs.logger
        self.configs = configs
        self.message_bus = message_bus

    def manage_operation_status(self, success, benefit_id, originator_pos_id, operation_description, benefit_applier, retry_quantity):
        # type: (bool, str, int, OperationDescription, BenefitAppliers, int) -> None

        if success:
            self.delete_pending_operation(benefit_id, originator_pos_id)
        else:
            retry_quantity += 1

            if retry_quantity > self.configs.max_retry_quantity:
                message = "The benefit has exceeded the max retry quantity. CouponId: {}; RetryQuantity: {}"
                self.logger.info(message.format(benefit_id, retry_quantity))

            self.insert_or_update_pending_operation(
                benefit_id,
                originator_pos_id,
                operation_description,
                OperationStatus.NOT_DONE,
                benefit_applier,
                retry_quantity
            )

    def find_pending_operations(self):
        # type: () -> Optional[List[BenefitControllerDto]]

        def inner_func(conn):
            # type: (Connection) -> Optional[List[BenefitControllerDto]]

            query = "SELECT * FROM PendingOperations WHERE OperationStatusId = '{}'".format(OperationStatus.NOT_DONE)
            cursor = conn.select(query)
            if not cursor.rows():
                return

            utc_now = iso8601.parse_date(datetime.utcnow().isoformat())
            acceptable_datetime_limit = timedelta(seconds=self.configs.retry_min_time_to_try_again)

            results = []
            for result in cursor:
                benefit = BenefitControllerDto()
                benefit.benefit_id = result.get_entry("BenefitId")
                benefit.originator_pos_id = result.get_entry("PosId")
                benefit.operation_description = int(result.get_entry("OperationDescriptionId"))
                benefit.benefit_applier = result.get_entry("BenefitApplier")
                benefit.retry_quantity = int(result.get_entry("RetryQuantity"))
                benefit.last_retry_date_utc = iso8601.parse_date(result.get_entry("LastRetryDateUTC"))

                exceed_acceptable_datetime_limit = benefit.last_retry_date_utc + acceptable_datetime_limit > utc_now
                exceed_acceptable_max_retry_quantity = benefit.retry_quantity > self.configs.max_retry_quantity
                if exceed_acceptable_datetime_limit or exceed_acceptable_max_retry_quantity:
                    continue

                results.append(benefit)

            return results

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def get_benefit_applier_by_db(self, benefit_id, pos_id):
        # type: (str, int) -> Optional[BenefitAppliers]

        def inner_func(conn):
            # type: (Connection) -> Optional[BenefitAppliers]

            query = "SELECT BenefitApplier FROM PendingOperations WHERE BenefitId = '{0}' AND PosId = '{1}'"
            cursor = conn.select(query.format(benefit_id, pos_id))
            if not cursor.rows():
                return

            return cursor.get_row(0).get_entry(0)

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def insert_or_update_pending_operation(self, benefit_id, originator_pos_id, operation_description, operation_status, benefit_applier, retry_quantity):
        # type: (str, int, OperationDescription, OperationStatus, BenefitAppliers, int) -> None

        def inner_func(conn):
            # type: (Connection) -> None

            last_retry_date = datetime.utcnow().isoformat()

            query = "INSERT OR REPLACE INTO PendingOperations(" \
                    "BenefitId, " \
                    "PosId, " \
                    "OperationDescriptionId, " \
                    "OperationStatusId, " \
                    "BenefitApplier, " \
                    "RetryQuantity, " \
                    "LastRetryDateUTC) " \
                    "VALUES ('{0}', '{1}', '{2}', '{3}', '{4}', '{5}', '{6}')"
            query = query.format(benefit_id,
                                 originator_pos_id,
                                 operation_description,
                                 operation_status,
                                 benefit_applier,
                                 retry_quantity,
                                 last_retry_date)
            conn.query(query)

        self.execute_with_transaction(inner_func, service_name=self.configs.persistence_name)

    def delete_pending_operation(self, benefit_id, originator_pos_id):
        # type: (str, int) -> None

        def inner_func(conn):
            # type: (Connection) -> None

            query = "DELETE FROM PendingOperations WHERE BenefitId = '{0}' AND PosId = '{1}'"\
                .format(benefit_id, originator_pos_id)
            conn.query(query)

        self.execute_with_transaction(inner_func, service_name=self.configs.persistence_name)



from abc import ABCMeta, abstractmethod

from model.repository import LogRepositoryModel
from model.services import EventDispatcherService


class BaseInteractor(object):
    __metaclass__ = ABCMeta

    def __init__(self, log_repository, event_dispatcher):
        # type: (LogRepositoryModel, EventDispatcherService) -> None
        self.log_repository = log_repository
        self.event_dispatcher = event_dispatcher

    @abstractmethod
    def execute(self, data):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from typing import List, Dict

from msgbus import MBEasyContext
from old_helper import BaseRepository
from persistence import Connection


class ProductRepository(BaseRepository):

    def __init__(self, mb_context):
        # type: (MBEasyContext) -> None

        super(ProductRepository, self).__init__(mb_context)
        self.all_part_code_jit_lines = None

    def get_all_part_code_jit_lines(self):
        # type: () -> Dict

        def func(conn):
            # type: (Connection) -> Dict

            cursor = conn.select(self._PartCodesOfJitLinesQuery)
            products_jit_lines = {}
            for row in cursor:
                products_jit_lines[int(row.get_entry("ProductCode"))] = row.get_entry("JITLines")

            return products_jit_lines

        if self.all_part_code_jit_lines is None:
            self.all_part_code_jit_lines = self.execute_with_connection(func)

        return self.all_part_code_jit_lines

    _PartCodesOfJitLinesQuery = """
        SELECT ProductCode, JITLines
        FROM Production 
        WHERE (Expiration IS NULL OR Expiration > DATETIME('now')) AND JITLInes IS NOT NULL
    """



import sysactions


def get_customer_info_config(pos_id):
    document = sysactions.get_cfg(pos_id).find_values("CustomerInfoRequest/Document") or []
    name = sysactions.get_cfg(pos_id).find_values("CustomerInfoRequest/Name") or []

    return {'document': document, 'name': name}



class ResyncRequest(object):
    def __init__(self, period=None, event_types=None, last_row_id=-1):
        self.period = period
        self.event_types = event_types
        self.last_row_id = last_row_id



"""
werkzeug.wrappers
~~~~~~~~~~~~~~~~~

The wrappers are simple request and response objects which you can
subclass to do whatever you want them to do.  The request object contains
the information transmitted by the client (webbrowser) and the response
object contains all the information sent back to the browser.

An important detail is that the request object is created with the WSGI
environ and will act as high-level proxy whereas the response object is an
actual WSGI application.

Like everything else in Werkzeug these objects will work correctly with
unicode data.  Incoming form data parsed by the response object will be
decoded into an unicode object if possible and if it makes sense.

:copyright: 2007 Pallets
:license: BSD-3-Clause
"""
from .accept import AcceptMixin
from .auth import AuthorizationMixin
from .auth import WWWAuthenticateMixin
from .base_request import BaseRequest
from .base_response import BaseResponse
from .common_descriptors import CommonRequestDescriptorsMixin
from .common_descriptors import CommonResponseDescriptorsMixin
from .etag import ETagRequestMixin
from .etag import ETagResponseMixin
from .request import PlainRequest
from .request import Request
from .request import StreamOnlyMixin
from .response import Response
from .response import ResponseStream
from .response import ResponseStreamMixin
from .user_agent import UserAgentMixin



# -*- coding: utf-8 -*-

import os
from messagehandler import MessageHandler
from helper import (
    config_logger,
    import_pydevd
)
from threading import Thread

import cfgtools
import pyscripts
from msgbus import MBEasyContext
from systools import sys_parsefilepath
from eventhandler import MaintenanceEventHandler
from events import MaintenanceProcessor, CleanDatabasesProcessor
from dbapitransactioncallback import DbApiTransactionCallback
from dbrepository import SqliteOrderRepository, SqliteFiscalRepository
from services import CommonService


def main():
    import_pydevd(os.environ["LOADERCFG"], 9128)

    required_services = "StoreWideConfig|Persistence|PosController"

    config_logger(os.environ["LOADERCFG"], 'Maintenance')
    config = cfgtools.read(os.environ["LOADERCFG"])

    mb_context = MBEasyContext("MaintenanceComp")
    pyscripts.mbcontext = mb_context

    start_hour = int(config.find_value("MaintenanceComponent.StartHour", default=5))
    end_hour = int(config.find_value("MaintenanceComponent.EndHour", default=6))

    comp_no = int(config.find_value("MaintenanceComponent.Number"))

    files_to_delete = config.find_values("MaintenanceComponent.DeleteFiles") or []
    dirs_to_delete = config.find_values("MaintenanceComponent.DeleteDirs") or []
    server_dirs_to_delete = config.find_values("MaintenanceComponent.DeleteServerDirs") or []

    common_service = CommonService(mb_context)
    maintenance_processor = MaintenanceProcessor(mb_context, common_service, comp_no, start_hour, end_hour)
    clean_databases_processor = None

    component_is_server = comp_no == 0
    if component_is_server:
        fiscal_path = sys_parsefilepath(config.find_value("MaintenanceComponent.XmlFiscalPath", default=""))
        parse_file_path = config.find_value("MaintenanceComponent.XmlFiscalEnviados", default="Enviados")
        fiscal_sent_path = sys_parsefilepath(parse_file_path)
        audit_path = sys_parsefilepath(config.find_value("MaintenanceComponent.AuditLogPath"))
        days_to_purge_audit_log = config.find_value("MaintenanceComponent.DaysToKeepAuditLogFiles", default=90)
        database_backups_to_keep = config.find_value("MaintenanceComponent.DatabaseBackupFilesToKeep", default=3)
        database_backups_to_keep = int(database_backups_to_keep)

        maintenance_processor.set_server_parameters(fiscal_path, audit_path, fiscal_sent_path, days_to_purge_audit_log)

        order_ddl_path = config.find_value("MaintenanceComponent.OrderSchemaDatabasePath", default='')
        fiscal_ddl_path = config.find_value("MaintenanceComponent.FiscalSchemaDatabasePath", default='')
        data_path = sys_parsefilepath("{HVDATADIR}")
        database_dir = os.path.join(data_path, "databases")
        backup_database_dir = os.path.join(data_path, "backups_maintenance/databases")
        order_schema_sql_path = os.path.join(data_path, order_ddl_path)
        fiscal_schema_sql_path = os.path.join(data_path, fiscal_ddl_path)

        db_api_callback = DbApiTransactionCallback()
        sqlite_order_repository = SqliteOrderRepository(
            common_service,
            database_dir,
            backup_database_dir,
            order_schema_sql_path,
            db_api_callback
        )

        sqlite_fiscal_repository = SqliteFiscalRepository(
            database_dir,
            backup_database_dir,
            fiscal_schema_sql_path,
            db_api_callback
        )

        clean_databases_processor = CleanDatabasesProcessor(
            mb_context,
            comp_no,
            database_backups_to_keep,
            sqlite_order_repository,
            sqlite_fiscal_repository,
            common_service
        )

    maintenance_event_handler = MaintenanceEventHandler(mb_context,
                                                        maintenance_processor,
                                                        clean_databases_processor)

    message_handler = MessageHandler(
        mb_context,
        "Maintenance%02d" % comp_no,
        "MaintenanceComp",
        required_services,
        maintenance_event_handler
    )

    if component_is_server:
        clean_databases_processor.set_server_parameters(maintenance_processor)
        delete_old_databases_thread = Thread(target=clean_databases_processor.clean_database_backups)
        delete_old_databases_thread.start()

    if files_to_delete:
        maintenance_processor.delete_files(files_to_delete)

    if dirs_to_delete or server_dirs_to_delete:
        maintenance_processor.delete_dirs(dirs_to_delete, server_dirs_to_delete)

    message_handler.subscribe_events(["CleanDatabaseBackups", "CleanDatabases", "VacuumDatabases"])
    message_handler.handle_events()



from ._ValidationUtil import \
    check_if_business_day_is_opened, \
    check_if_operator_is_logged, \
    check_if_order_is_in_progress, \
    check_if_order_is_totaled, \
    check_if_order_is_in_progress_or_totaled, \
    check_if_pos_id_is_valid, \
    check_if_payment_type_is_valid, \
    check_if_payment_amount_is_valid, \
    check_if_due_amount_is_zero, \
    check_business_period, \
    validate_api_key, \
    get_pos_id_from_pos_mapper, \
    get_abstracted_pos_id_from_pos_mapper




from _TefCancellationBodyDto import TefCancellationBodyDto
from _TefCancellationGeneratorDto import TefCancellationGeneratorDto
from _TefCancellationHeaderDto import TefCancellationHeaderDto

__all__ = ["TefCancellationBodyDto", "TefCancellationGeneratorDto", "TefCancellationHeaderDto"]



from get_authorization import get_authorization
from list_users import list_opened_users, check_valid_user, is_user_already_logged, set_context
from get_drawer_amount import get_drawer_amount
from get_drawer_initial_amount import get_drawer_initial_amount
from get_pos_user import get_pos_users, get_opened_pos_user
from get_working_type import get_working_type
from get_operator_info import get_operator_info
from get_operator_level import get_user_level
from get_operator_name import get_operator_name
from get_payments_amount_by_type import get_payments_amount_by_type
from cash_out_functions import do_set_drawer_status, check_sangria, is_sangria_enable
from check_current_order import check_active_order
from get_ruptured_products import get_ruptured_products_from_remote_order
from delivery_functions import is_delivery_order
from pos_type import set_pos_type, get_pos_types



# -*- coding: utf-8 -*-
import json
import time

from posot import OrderTaker
from typing import Dict, Optional, List, Any
from threading import Lock
from xml.etree.ElementTree import Element
from xml.etree import cElementTree as eTree

from sysactions import get_model, get_posot
from messagebus import MessageBus

from application.service import OrderService, BenefitList
from application.util import DictionaryOperations
from application.validator import BenefitValidator
from application.model import Configurations
from application.model.benefit import ActionResult, Benefit, OrderBenefitInfo
from application.model.conditions import Restrictions
from application.model.transactions import AddBenefitToApplyUnitOfWork, VerifyAndRemoveBenefitsUnitOfWork
from application.model.exceptions import (
    OrderValidationException,
    BenefitNotFound,
    InvalidMessageContentException,
    DiscountNotFound,
    CannotRemoveItemsInATotaledOrder,
)
from application.repository import NDiscountRepository

from ._ActionService import ActionService


class BenefitService(object):
    def __init__(self, configs, repository, message_bus, order_service):
        # type: (Configurations, NDiscountRepository, MessageBus, OrderService) -> None

        self.configs = configs
        self.repository = repository
        self.message_bus = message_bus
        self.order_service = order_service
        self.logger = configs.logger

        self.product_groups = self.repository.get_all_product_groups()
        self.action_service = ActionService(order_service, self.product_groups, self.logger)
        self.automatic_benefits = self._get_benefit_types()

        # TODO: Check if we really need a Lock
        self.discount_lock = {}

    def get_unit_of_work(
            self,
            pos_id,                     # type: int
            raw_benefit,                # type: Dict
            order_picture=None,         # type: Optional[Element]
            benefit_applier=None,       # type: Optional[str]
            tender_id=None,             # type: Optional[str]
            table_id=None,              # type: Optional[str]
            user_auth_level=False       # type: Optional[bool]
    ):
        # type: (...) -> AddBenefitToApplyUnitOfWork

        benefit = Benefit(raw_benefit)
        uow = AddBenefitToApplyUnitOfWork(
            self.logger,
            pos_id,
            benefit,
            benefit_applier,
            tender_id,
            order_picture,
            self.order_service,
            self.repository,
            self.product_groups,
            table_id=table_id,
            user_auth_level=user_auth_level
        )

        return uow

    def apply_automatic_benefits_on_table(
            self,
            pos_id,         # type: int
            table_id,       # type: str
            is_tab=False    # type: Optional[bool]
    ):
        # type: (...) -> Optional[str]

        if pos_id not in self.discount_lock:
            self.discount_lock[pos_id] = Lock()
        elif self.discount_lock[pos_id].locked():
            return

        with self.discount_lock[pos_id]:
            try:
                discount_type = "ON_TAB_OPEN" if is_tab else "ON_TABLE_OPEN"
                if len(self.automatic_benefits) > 0:
                    if self.automatic_benefits.get(discount_type) is not None:
                        return self._apply_table_benefits(
                            pos_id=pos_id,
                            table_id=table_id,
                            discount_type=discount_type
                        )
                    else:
                        return

            except Exception as _:
                self.logger.exception("Error applying automatic discounts on POS {}".format(str(pos_id)))
                raise

    def apply_automatic_benefits_on_tab(
            self,
            pos_id,     # type: int
            table_id    # type: str
    ):
        # type: (...) -> Optional[str]

        if pos_id not in self.discount_lock:
            self.discount_lock[pos_id] = Lock()
        elif self.discount_lock[pos_id].locked():
            return

        with self.discount_lock[pos_id]:
            try:
                if len(self.automatic_benefits) > 0:
                    if self.automatic_benefits.get("ON_TAB_OPEN") is not None:
                        return self._apply_tab_benefits(
                            pos_id=pos_id,
                            table_id=table_id
                        )
                    else:
                        return

            except Exception as _:
                self.logger.exception("Error applying automatic discounts on POS {}".format(str(pos_id)))
                raise

    def apply_automatic_benefits_on_order(self, pos_id, on_total=False):
        # type: (int, bool) -> Optional[eTree]

        if pos_id not in self.discount_lock:
            self.discount_lock[pos_id] = Lock()
        elif self.discount_lock[pos_id].locked():
            return

        order_picture = None
        with self.discount_lock[pos_id]:
            try:
                if len(self.automatic_benefits) > 0:
                    order_picture = self.order_service.get_order_picture(pos_id)
                    order_state = order_picture.get("state")

                    if not on_total:
                        if order_state == "IN_PROGRESS" and self.automatic_benefits.get("ON_SALE") is not None:
                            for benefit in self.automatic_benefits["ON_SALE"]:
                                updated_order = self._apply_automatic_benefit(pos_id, benefit, order_picture)
                                order_picture = updated_order if updated_order is not None else order_picture
                    elif self.automatic_benefits.get("ON_TOTAL") is not None:
                        for benefit in self.automatic_benefits["ON_TOTAL"]:
                            updated_order = self._apply_automatic_benefit(pos_id, benefit, order_picture, on_total)
                            order_picture = updated_order if updated_order is not None else order_picture

            except Exception as _:
                self.logger.exception("Error applying automatic discounts on POS {}".format(str(pos_id)))
                raise

        return order_picture

    def apply_manual_pending_benefits(self, pos_id, order_picture=None):
        # type: (int, Optional[eTree]) -> None

        model = get_model(pos_id)
        if self._is_order_taker(model):
            self.logger.info("The current model is order taker. Ignoring benefit apply for POS: {}".format(pos_id))
            return

        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
        applied_benefits_keys = [b["benefit_key"] for b in applied_benefits]
        benefit_list = self.order_service.get_benefit_list_from_order_picture(pos_id, order_picture)

        if "manualBenefitOnTotalSequence" not in benefit_list:
            return

        try:
            manual_benefits = benefit_list["manualBenefitOnTotalSequence"] or []
            for item in manual_benefits:
                key, benefit = item.items()[0]
                if key not in applied_benefits_keys:
                    uow = self.get_unit_of_work(pos_id, benefit)
                    self.apply_benefit(uow, is_manual=True, remove_manual_key=key)
        except OrderValidationException as _:
            pass
        except Exception as _:
            exception_message = "Error apply automatic benefit for order {} on POS {}"
            self.logger.exception(exception_message.format(order_picture.get("orderId", ""), str(pos_id)))

    def apply_pending_benefits(self, pos_id, order_picture=None):
        # type: (int, Optional[eTree]) -> None

        model = get_model(pos_id)
        if self._is_order_taker(model):
            self.logger.info("The current model is order taker. Ignoring benefit apply for POS: {}".format(pos_id))
            return

        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id)
        order_id = order_picture.get("orderId")
        order_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)

        if not order_benefits:
            self.logger.info("No benefits to orderId: {}; posId: {}".format(order_id, pos_id))

        self.logger.info("Trying to apply benefits to orderId: {}; posId: {}".format(order_id, pos_id))
        for order_benefit_json in order_benefits:
            benefit_order_info = OrderBenefitInfo.from_json(order_benefit_json)
            if benefit_order_info.benefit_applied:
                continue

            self._apply_pending_benefit(pos_id, benefit_order_info)

    def apply_benefit(self, uow, validate_benefits=True, on_total=False, is_manual=False, remove_manual_key=''):
        # type: (AddBenefitToApplyUnitOfWork, bool, bool, bool, str) -> None

        try:
            message = "Adding benefit to orderId: {}; posId: {}; benefitName: {}"
            self.logger.info(message.format(uow.order.id, uow.pos_id, uow.benefit.name))

            order_picture = uow.order.order_picture
            if order_picture is None:
                order_picture = self.order_service.get_order_picture(uow.pos_id)

            applied_benefits = self.order_service.get_applied_benefits_from_order_picture(uow.pos_id, order_picture)
            burnt_items = self.order_service.get_burnt_items(applied_benefits)

            if validate_benefits:
                BenefitValidator.validate_benefits(uow, applied_benefits, burnt_items, self.product_groups)

            benefit_key = self._get_benefit_key(uow.benefit_id)
            uow.set_benefit_key(benefit_key)
            action_result = ActionResult([], {}, [], {}, True, False)
            already_burnt = False
            for action in uow.benefit.actions:
                action.times_to_apply =\
                    uow.benefit.times_to_apply or (uow.benefit_order_info and uow.benefit_order_info.times_to_apply)
                restrictions = Restrictions(uow.pos_id, order_picture, uow.benefit, self.repository, self.order_service)
                result = self.action_service.apply_action(action, burnt_items, uow, uow.pos_ot, restrictions, on_total)
                action_result += result
                items_to_burn = result.items_to_burn
                if len(items_to_burn) > 0 and not already_burnt:
                    updated_items_to_burn = DictionaryOperations.unmerge_counting_dict(items_to_burn, burnt_items)
                    action_result.items_to_burn = updated_items_to_burn
                    already_burnt = True

            if action_result:
                self._save_benefits_list(uow, uow.pos_ot, action_result, is_manual, remove_manual_key)
                uow.update_order()

                message = "Benefit applied successfully to orderId: {}; posId: {}; benefitName: {};"
                self.logger.info(message.format(uow.order.id, uow.pos_id, uow.benefit.name))
            else:
                message = "Benefit not applied to orderId: {}; posId: {}; benefitName: {};"
                self.logger.info(message.format(uow.order.id, uow.pos_id, uow.benefit.name))
        except OrderValidationException as ex:
            message = "Invalid benefit to orderId: {}; posId: {}; benefitName: {}; validation: {}"
            self.logger.info(message.format(uow.order.id, uow.pos_id, uow.benefit.name, ex.message))
            raise

    def validate_benefits(self, pos_id, order_picture=None, product_groups=None):
        # type: (int, Optional[eTree], Dict) -> eTree

        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id)

        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
        if not applied_benefits:
            return order_picture

        order = self.order_service.get_order(pos_id, order_picture=order_picture)
        uow = VerifyAndRemoveBenefitsUnitOfWork(pos_id, order)

        log_message = "Validating used benefits from orderId: {}; posId: {}"
        self.logger.info(log_message.format(uow.order.id, uow.pos_id))

        burnt_items = DictionaryOperations.merge_counting_dicts([
            applied_benefit.get("burnt_items", {}) for applied_benefit in applied_benefits
        ])

        removed_benefit = False
        for applied_benefit in applied_benefits:
            benefit = Benefit(applied_benefit.get("benefit"))
            benefit_key = applied_benefit.get("benefit_key")
            benefit_burnt_items = applied_benefit.get("burnt_items", {})
            burnt_items_for_benefit = DictionaryOperations.unmerge_counting_dict(burnt_items, benefit_burnt_items)
            valid_benefit = self._validate_benefit(benefit_key, benefit, uow, burnt_items_for_benefit, product_groups)
            if valid_benefit is False:
                removed_benefit = True

        if removed_benefit:
            order = self.order_service.get_order(pos_id)
            if sum([x.qty for x in order.sale_lines]) == 0:
                self.order_service.void_order(pos_id, uow.pos_ot)

            return order.order_picture

        return order_picture

    def remove_benefits(self, pos_id, benefit_keys, order_picture=None, delete_manual=True):
        # type: (int, List, Optional[eTree], bool) -> None

        model = get_model(pos_id)
        pos_ot = get_posot(model)
        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id)

        order_id = order_picture.get("orderId")
        log_message = "Trying to remove benefits to orderId: {}; posId: {}; benefitIds: {}"
        self.logger.info(log_message.format(order_id, pos_id, benefit_keys))

        if benefit_keys is None:
            self.logger.info("BenefitIds not provided. Removing all order benefits")
            all_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
            benefit_keys = [x.get("benefit_key") for x in all_benefits]

        self.remove_benefit(pos_id, pos_ot, benefit_keys, order_picture, delete_manual)

    def remove_benefit(self, pos_id, pos_ot, benefit_keys, order_picture=None, delete_manual=True):
        # type: (int, eTree, List[str], Optional[eTree], Optional[bool]) -> None

        self._remove_benefit_by_key(pos_id, pos_ot, benefit_keys, order_picture, delete_manual)

    def unset_applied_benefit(self, pos_id, order_state):
        # type: (int, List) -> None

        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id)
        benefits = []
        tenders_remove = []
        discounts_remove = []
        item_amounts_remove = {}
        for applied_benefit in applied_benefits:
            benefit_applied = applied_benefit.get("benefit_applied")
            applied_order_state = applied_benefit.get("order_state")
            if benefit_applied and applied_order_state in order_state:
                item_amounts_remove.update(applied_benefit.get("discount_item_values", {}))
                tenders_remove.extend(applied_benefit.get("order_tenders_ids", "[]"))
                discounts_remove.extend(applied_benefit.get("used_discount_ids", "[]"))
                applied_benefit["benefit_applied"] = False
                applied_benefit["order_tenders_ids"] = []
                applied_benefit["used_discount_ids"] = []

            benefits.append(applied_benefit)

        model = get_model(pos_id)
        pos_ot = get_posot(model)
        self.order_service.clear_discounts(pos_ot, discounts_remove)
        self.order_service.clear_tenders(pos_id, pos_ot, tenders_remove)

        self._set_benefits_to_apply(
            pos_id,
            benefits,
            pos_ot,
            [],
            item_amounts_remove,
            tenders_remove,
            discounts_remove,
            delete_manual=False
        )

    def remove_invalid_benefits(self, pos_id, order_picture=None):
        # type: (int, Optional[eTree]) -> eTree

        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id)
        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
        if not applied_benefits:
            return order_picture

        order = self.order_service.get_order(pos_id, order_picture=order_picture)
        benefits_remove = []
        for applied_benefit in applied_benefits:
            benefit_order_info = OrderBenefitInfo.from_json(applied_benefit)
            if not BenefitValidator.validate_added_sale_line_benefit(benefit_order_info, order):
                benefits_remove.append(benefit_order_info.benefit_key)

        if len(benefits_remove) > 0:
            self.remove_benefits(pos_id, benefits_remove, order_picture)

        return order_picture

    def _remove_benefit_by_key(self, pos_id, pos_ot, benefit_keys, order_picture=None, delete_manual=True):
        # type: (int, OrderTaker, List[str], Optional[eTree], Optional[bool]) -> None

        benefits_to_remove = self._get_applied_benefits_by_key(pos_id, benefit_keys, order_picture)

        self.check_if_benefit_is_removable(
            pos_id=pos_id,
            benefits_to_remove=benefits_to_remove,
            order_picture=order_picture
        )

        added_sale_lines = []
        used_discount_ids = []
        used_order_tender_ids = []

        for benefit_to_remove in benefits_to_remove:
            added_sale_lines.extend(benefit_to_remove.get("added_sale_lines", []))
            used_discount_ids.extend(benefit_to_remove.get("used_discount_ids", []))
            used_order_tender_ids.extend(benefit_to_remove.get("order_tenders_ids", []))

        self.order_service.void_line(pos_id, pos_ot, added_sale_lines)
        self.order_service.clear_discounts(pos_ot, used_discount_ids)
        self.order_service.clear_tenders(pos_id, pos_ot, used_order_tender_ids)

        self._delete_benefit(benefit_keys, pos_ot, pos_id, order_picture, delete_manual)

    def check_if_benefit_is_removable(
        self,
        pos_id,                 # type: str
        benefits_to_remove,     # type: List[Dict]
        order_picture=None      # type: Optional[eTree.XML]
    ):
        # type: (...) -> None

        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id)

        state = order_picture.get("state")

        if state == u"TOTALED":
            for benefit in benefits_to_remove:
                benefit_action = benefit.get("benefit")
                actions = benefit_action.get("actions")
                for action in actions.values():
                    if action.get("itemsToAdd"):
                        raise CannotRemoveItemsInATotaledOrder()

    def _save_benefits_list(self, uow, pos_ot, action_result, is_manual=False, remove_manual_key=''):
        # type: (AddBenefitToApplyUnitOfWork, OrderTaker, Optional[ActionResult], Optional[bool], Optional[str]) -> None

        order_picture = uow.order.order_picture
        old_benefits = self.order_service.get_applied_benefits_from_order_picture(uow.pos_id, order_picture)
        benefit_key = uow.benefit_key

        if uow.benefit_order_info is None:
            order_state = order_picture.get("state")
            str_data = self._get_new_benefit(uow, action_result, old_benefits, benefit_key, order_state)
        else:
            benefit_key = uow.benefit_order_info.benefit_key
            str_data = self._update_benefit(uow, action_result)

        discount_ids = action_result.discounts_id if not uow.benefit.tip_included else []
        tenders_ids = action_result.order_tenders
        item_values = action_result.discount_item_values
        benefit_list = self._get_benefit_list_to_add(
            uow, str_data, benefit_key, item_values, tenders_ids, discount_ids, is_manual, remove_manual_key
        )
        self.order_service.save_benefit_list_custom_property(pos_ot, benefit_list)

    def _validate_benefit(self, benefit_key, benefit, uow, burnt_items_for_benefit, product_groups):
        # type: (str, Benefit, VerifyAndRemoveBenefitsUnitOfWork, Dict, Dict) -> bool

        try:
            BenefitValidator.validate_benefit_order_info(benefit, uow.order, burnt_items_for_benefit, product_groups)
            return True
        except OrderValidationException:
            self.remove_benefit(uow.pos_id, uow.pos_ot, [benefit_key])
            return False

    @staticmethod
    def _get_new_benefit(uow, action_result, old_benefits, benefit_key, order_state):
        # type: (AddBenefitToApplyUnitOfWork, Optional[ActionResult], List, str, str) -> str

        benefit_total_value = BenefitService._sum_discounts(uow.promotion_tenders, action_result.discount_item_values)
        new_benefits = [dict(
            benefit=uow.raw_benefit,
            benefit_id=uow.raw_benefit.get("name"),
            benefit_total_value=benefit_total_value,
            burnt_items=action_result.items_to_burn,
            added_sale_lines=uow.added_sale_lines,
            promotion_tender=uow.promotion_tenders,
            benefit_applier=uow.benefit_applier,
            used_discount_ids=action_result.discounts_id,
            order_tenders_ids=action_result.order_tenders,
            discount_item_values=action_result.discount_item_values,
            benefit_applied=action_result.benefit_applied,
            tender_id=uow.tender_id,
            order_state=order_state,
            benefit_key=benefit_key,
            times_to_apply=uow.benefit.times_to_apply)
        ]

        return json.dumps(old_benefits + new_benefits)

    def _update_benefit(self, uow, action_result):
        # type: (AddBenefitToApplyUnitOfWork, Optional[ActionResult]) -> str

        old_benefits = self.order_service.get_benefit_list_from_order_picture(uow.pos_id)
        benefit_list = json.loads(old_benefits.get("benefitList", "[]"))
        list_benefit = []
        for benefit in benefit_list:
            if benefit["benefit_key"] == uow.benefit_order_info.benefit_key:
                benefit["discount_item_values"] = action_result.discount_item_values
                benefit["used_discount_ids"] = action_result.discounts_id
                benefit["order_tenders_ids"] = action_result.order_tenders
                benefit["benefit_applied"] = action_result.benefit_applied
                benefit["tender_id"] = uow.tender_id

            list_benefit.append(benefit)
        return json.dumps(list_benefit)

    def _delete_benefit(self, benefit_keys, pos_ot, pos_id, order_picture=None, delete_manual=True):
        # type: (List[str], OrderTaker, int, Optional[eTree], Optional[bool]) -> None

        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
        after_deletion = []
        keys_remove = []
        item_amounts_remove = {}
        tenders_remove = []
        discounts_remove = []

        for raw_benefit in applied_benefits:
            benefit_json_key = raw_benefit.get("benefit_key")
            if benefit_json_key not in benefit_keys:
                after_deletion.append(raw_benefit)
                continue

            keys_remove.append(raw_benefit.get("benefit_key"))
            item_amounts_remove.update(raw_benefit.get("discount_item_values"))
            tenders_remove.extend(raw_benefit.get("order_tenders_ids"))
            discounts_remove.extend(raw_benefit.get("used_discount_ids"))

        self._set_benefits_to_apply(
            pos_id,
            after_deletion,
            pos_ot,
            keys_remove,
            item_amounts_remove,
            tenders_remove,
            discounts_remove,
            order_picture,
            delete_manual
        )

    def _set_benefits_to_apply(
            self,
            pos_id,                 # type: int
            benefits,               # type: List[str]
            pos_ot,                 # type: OrderTaker
            keys_remove,            # type: List[str]
            item_amounts_remove,    # type: Dict
            tenders_remove,         # type: List[int]
            discounts_remove,       # type: List[int]
            order_picture=None,     # type: Optional[eTree]
            delete_manual=True      # type: Optional[bool]
    ):
        # type: (...) -> None

        if not benefits:
            pos_ot.setOrderCustomProperty("BENEFIT_LIST", "clear")
            return

        data = json.dumps(benefits)
        benefit_list = self._get_benefit_list_to_remove(
            pos_id,
            data,
            keys_remove,
            item_amounts_remove,
            tenders_remove,
            discounts_remove,
            order_picture,
            delete_manual
        )
        self.order_service.save_benefit_list_custom_property(pos_ot, benefit_list)

    def _get_applied_benefits_by_key(self, pos_id, benefit_keys, order_picture=None):
        # type: (int, List[str], Optional[eTree]) -> List

        matching_benefits = []
        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
        for applied_benefit in applied_benefits:
            applied_benefit_key = applied_benefit.get("benefit_key")
            if applied_benefit_key in benefit_keys:
                matching_benefits.append(applied_benefit)

        if not matching_benefits:
            raise DiscountNotFound()

        return matching_benefits

    def _get_benefit_types(self):
        response = {}
        automatic_benefits = self.repository.get_all_automatic_benefits()
        for benefit in automatic_benefits:
            benefit_json = json.loads(benefit)
            benefit = benefit_json.get("benefit")
            event = benefit.get("automatic")

            if event not in response:
                response[event] = []

            response[event].append(benefit)

        return response

    def _apply_automatic_benefit(self, pos_id, benefit, order_picture=None, on_total=False, table_id=None):
        # type: (int, Dict, Element, bool, str) -> Optional[Element]

        order_id = None
        uow = None
        try:
            uow = self.get_unit_of_work(pos_id, benefit, order_picture, table_id=table_id)
            order_id = uow.order.id
            self.apply_benefit(uow, on_total=on_total)
            return uow.order.order_picture
        except OrderValidationException as _:
            pass
        except Exception as _:
            self.logger.exception(self._get_error_message(order_id, uow.benefit))

    def _apply_pending_benefit(self, pos_id, benefit_order_info):
        # type: (int, OrderBenefitInfo) -> None

        order_id = None
        benefit = None
        try:
            benefit = benefit_order_info.benefit
            uow = self.get_unit_of_work(pos_id, benefit.raw)
            uow.benefit_order_info = benefit_order_info
            order_id = uow.order.id
            self.apply_benefit(uow, validate_benefits=False)

        except OrderValidationException as _:
            pass
        except Exception as _:
            self.logger.exception(self._get_error_message(order_id, benefit))

    @staticmethod
    def _get_error_message(
            order_id,    # type: int
            benefit      # type: Benefit
    ):
        # type: (...) -> str

        return "Error applying benefit for order {} - Benefit name: {}".format(str(order_id), benefit.name)

    @staticmethod
    def _is_order_taker(model):
        # type: (eTree) -> bool

        working_mode = model.find(".//WorkingMode")
        pos_function = working_mode.get("posFunction")

        if pos_function is not None and pos_function == "OT":
            return True
        return False

    def _get_benefit_list_to_add(
            self,
            uow,                # type: AddBenefitToApplyUnitOfWork
            data,               # type: str
            benefit_key,        # type: Any[str, List[str]]
            item_values,        # type: Dict
            order_tenders_ids,  # type: List
            discount_ids,       # type: List
            is_manual,          # type: bool
            remove_manual_key   # type: str
    ):
        # type: (...) -> BenefitList

        order_picture = uow.order.order_picture
        benefits_pre_total = self._get_benefits_pre_total(uow.pos_id, benefit_key, order_picture)
        manual_benefits_on_total = self._get_manual_benefits_on_total(uow, order_picture, is_manual, remove_manual_key)
        discount_item_values = self._get_discount_item_values(uow.pos_id, item_values, order_picture)
        discount_excluded = self._get_discount_type_excluded(uow.pos_id, discount_ids, order_picture)
        tender_excluded = self._get_tender_type_excluded(uow.pos_id, order_tenders_ids, order_picture)

        return BenefitList(
            benefits_pre_total, manual_benefits_on_total, discount_item_values, discount_excluded, tender_excluded, data
        )

    def _get_benefit_list_to_remove(
            self,
            pos_id,                 # type: int
            data,                   # type: str
            benefit_key,            # type: Any[str, List[str]]
            item_values,            # type: Dict
            order_tenders_ids,      # type: List[int]
            discounts_remove,       # type: List[int]
            order_picture=None,     # type: Optional[eTree]
            delete_manual=True      # type: Optional[bool]
    ):
        # type: (...) -> BenefitList

        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id)

        benefits_pre_total = self._get_benefits_pre_total(pos_id, benefit_key, order_picture)
        manual_benefits_on_total = self._get_removed_manual_benefits(pos_id, benefit_key, order_picture, delete_manual)
        discount_item_values = self._remove_properties(pos_id, item_values, "discountItemValues", order_picture, True)
        discount_excluded = self._remove_properties(pos_id, discounts_remove, "discountIdTipExcluded", order_picture)
        tender_excluded = self._remove_properties(pos_id, order_tenders_ids, "tenderIdTipExcluded", order_picture)

        return BenefitList(
            benefits_pre_total, manual_benefits_on_total, discount_item_values, discount_excluded, tender_excluded, data
        )

    def _get_benefits_pre_total(self, pos_id, benefit_key, order_picture=None):
        # type: (int, [str, List[str]], Element) -> List

        benefits = self.order_service.get_benefit_list_from_order_picture(pos_id, order_picture)
        if isinstance(benefit_key, List) and len(benefit_key) > 0 and len(benefits) > 0:
            return self._remove_list_properties_benefits(benefits, benefit_key, "benefitPreTotalSequence")

        if len(benefits) == 0:
            return [benefit_key]

        if order_picture.get("state") == "TOTALED":
            return benefits.get("benefitPreTotalSequence")

        pre_total_sequence = benefits.get("benefitPreTotalSequence")
        if len(benefit_key) > 0:
            pre_total_sequence.append(benefit_key)
        return pre_total_sequence

    def _get_manual_benefits_on_total(
            self,
            uow,                    # type: AddBenefitToApplyUnitOfWork
            order_picture=None,     # type: Optional[eTree.XML]
            is_manual=False,        # type: Optional[bool]
            remove_manual_key=''    # type: Optional[str]
    ):
        # type: (...) -> List

        benefits = self.order_service.get_benefit_list_from_order_picture(uow.pos_id, order_picture)
        manual_sequence = benefits.get("manualBenefitOnTotalSequence", [])
        if order_picture.get("state") != "TOTALED":
            return manual_sequence

        new_manual_sequence = []
        for item in manual_sequence:
            key, benefit = item.items()[0]
            if key != remove_manual_key:
                new_manual_sequence.append({key: benefit})

        benefit_key = uow.benefit_key
        if len(benefit_key) > 0 and is_manual:
            raw_benefit = uow.benefit.raw
            new_manual_sequence.append({benefit_key: raw_benefit})

        return new_manual_sequence

    def _get_removed_manual_benefits(
            self,
            pos_id,                 # type: int
            benefit_keys,           # type: List[str]
            order_picture,          # type: eTree.XML
            delete_manual=True      # type: Optional[bool]
    ):
        # type: (...) -> List[Dict]

        benefits = self.order_service.get_benefit_list_from_order_picture(pos_id, order_picture)

        if len(benefits) == 0:
            return []

        manual_benefits = benefits.get("manualBenefitOnTotalSequence", [])
        filtered_manual_benefits = []
        for benefit in manual_benefits:
            key, value = benefit.items()[0]
            if key not in benefit_keys or not delete_manual:
                filtered_manual_benefits.append({key: value})

        return filtered_manual_benefits

    def _get_discount_item_values(self, pos_id, item_discount_values, order_picture=None):
        # type: (int, Dict, Element) -> List

        benefits = self.order_service.get_benefit_list_from_order_picture(pos_id, order_picture)
        return self._get_dict_data_remove(benefits, item_discount_values, "discountItemValues")

    def _get_discount_type_excluded(self, pos_id, discount_ids, order_picture=None):
        # type: (int, List[str], Element) -> List

        benefits = self.order_service.get_benefit_list_from_order_picture(pos_id, order_picture)
        return self._get_list_data_remove(benefits, discount_ids, "discountIdTipExcluded")

    def _get_tender_type_excluded(self, pos_id, order_tenders_ids, order_picture=None):
        # type: (int, List[str], Element) -> List

        benefits = self.order_service.get_benefit_list_from_order_picture(pos_id, order_picture)
        if len(benefits) == 0 or len(order_tenders_ids) == 0:
            return order_tenders_ids

        return self._get_list_data_remove(benefits, order_tenders_ids, "tenderIdTipExcluded")

    def _remove_properties(self, pos_id, data, property_name, order_picture, is_dict_data=False):
        # type: (int, List[int], str, Element, bool) -> List

        benefits = self.order_service.get_benefit_list_from_order_picture(pos_id, order_picture)

        if is_dict_data:
            return self._remove_dict_properties_benefits(benefits, data, property_name)
        return self._remove_list_properties_benefits(benefits, data, property_name)

    @staticmethod
    def _get_benefit_key(benefit_id):
        # type: (str) -> str

        return "{}_{}".format(benefit_id, int(round(time.time() * 1000)))

    @staticmethod
    def _get_dict_data_remove(benefits, dict_data, property_name):
        # type: (List, Dict, str) -> Dict

        ids = benefits.get(property_name, {})
        for data_id in dict_data.keys():
            if data_id not in ids.keys():
                ids[data_id] = dict_data[data_id]

        return ids

    @staticmethod
    def _get_list_data_remove(benefits, list_data, property_name):
        # type: (Dict, List, str) -> List

        if len(benefits) == 0:
            return []

        ids = benefits.get(property_name, "[]")
        for data_id in list_data:
            if data_id not in ids:
                ids.append(data_id)

        return ids

    @staticmethod
    def _remove_list_properties_benefits(benefits, list_remove, property_name):
        # type: (Dict, List, str) -> List

        result = []
        keys = benefits.get(property_name, "[]")
        for key in keys:
            if key not in list_remove:
                result.append(key)

        return result

    @staticmethod
    def _remove_dict_properties_benefits(benefits, dict_remove, property_name):
        # type: (List, Dict, str) -> List

        result = {}
        body = benefits.get(property_name, {})
        for key in body.keys():
            if key not in dict_remove.keys():
                result[key] = body[key]

        return result

    def _apply_table_benefits(
            self,
            pos_id,         # type: int
            table_id,       # type: str
            discount_type   # type: str
    ):
        # type: (...) -> Optional[str]

        for benefit in self.automatic_benefits[discount_type]:
            self._apply_automatic_benefit(
                pos_id=pos_id,
                benefit=benefit,
                table_id=table_id
            )

        order = self.order_service.get_order(pos_id)
        order_id = order.id if order else ''
        if sum([x.qty for x in order.sale_lines]) == 0:
            model = get_model(pos_id)
            pos_ot = get_posot(model)
            self.order_service.void_order(pos_id, pos_ot)
            return
        return str(order_id)

    @staticmethod
    def _sum_discount_items(discount_item_values):
        # type: (Dict) -> float

        total_discount = 0
        for key, value in discount_item_values.items():
            if isinstance(value, float):
                total_discount += value
                continue

            total_discount += BenefitService._sum_discount_items(value)

        return total_discount

    @staticmethod
    def _sum_discounts(promotion_tender, discount_item_values):
        # type: (Dict, Dict) -> float

        sum_promotion_tender = sum([float(x.get("amount")) for x in promotion_tender])
        sum_discount_item_values = BenefitService._sum_discount_items(discount_item_values)

        return sum_promotion_tender + sum_discount_item_values



# -*- coding: utf-8 -*-

import ctypes
import os
import pickle
import platform

from cache import CacheManager, PeriodBasedExpiration
from persistence import Driver


class FingerPrintReader(object):
    def __init__(self):
        pass

    def enroll_user(self, user_id, pos_id, replace):
        pass

    def authenticate(self, pos_id):
        pass


class EnrolledUsersRepository(object):
    def __init__(self):
        pass

    def load_all_enrolled_users(self):
        raise NotImplementedError()

    def save_enrolled_user(self, enrolled_user, pos_id):
        raise NotImplementedError()

    def remove_enrolled_user(self, enrolled_user, pos_id):
        raise NotImplementedError()


class ObjectFileEnrolledUsersRepository(EnrolledUsersRepository):
    FINGER_PRINT_DATA_FILE = "finger_print_data_file.obj"

    def __init__(self):
        super(ObjectFileEnrolledUsersRepository, self).__init__()
        self.enrolled_users = []

    def load_all_enrolled_users(self):
        if not self.enrolled_users:
            if os.path.isfile(ObjectFileEnrolledUsersRepository.FINGER_PRINT_DATA_FILE):
                with open(ObjectFileEnrolledUsersRepository.FINGER_PRINT_DATA_FILE, "rb") as file:
                    self.enrolled_users = pickle.load(file)

        return self.enrolled_users

    def save_enrolled_user(self, enrolled_user, _):
        self.enrolled_users.append(enrolled_user)

        with open(ObjectFileEnrolledUsersRepository.FINGER_PRINT_DATA_FILE, "wb") as file:
            pickle.dump(self.enrolled_users, file)

    def remove_enrolled_user(self, enrolled_user, _):
        for already_enrolled_user in self.enrolled_users:
            if already_enrolled_user.user_id == enrolled_user.user_id:
                self.enrolled_users.remove(already_enrolled_user)
                break

        with open(ObjectFileEnrolledUsersRepository.FINGER_PRINT_DATA_FILE, "wb") as file:
            pickle.dump(self.enrolled_users, file)


class DatabaseEnrolledUsersRepository(EnrolledUsersRepository):
    def __init__(self, mbcontext, enrolled_users_expiration):
        # type: (MBEasyContext) -> None
        super(DatabaseEnrolledUsersRepository, self).__init__()
        self.mbcontext = mbcontext
        self.enrolled_users = []
        self.enrolled_users_cache = CacheManager(PeriodBasedExpiration(enrolled_users_expiration), threaded=False)

    def load_all_enrolled_users(self):
        return self.enrolled_users_cache.get_cached_object(self._renew_enrolled_users_cache)

    def save_enrolled_user(self, enrolled_user, pos_id):
        conn = None
        try:
            conn = Driver().open(self.mbcontext, dbname=str(pos_id))
            conn.query("update Users set FingerPrintMinutiae = x'%s' where UserId = %s" % (enrolled_user.fmd_buffer.encode("hex"), enrolled_user.user_id))
        finally:
            if conn:
                conn.close()

        self.enrolled_users.append(enrolled_user)

    def remove_enrolled_user(self, enrolled_user, pos_id):
        conn = None
        try:
            conn = Driver().open(self.mbcontext, dbname=str(pos_id))
            conn.query("update Users set FingerPrintMinutiae = NULL where UserId = %s" % enrolled_user.user_id)
        finally:
            if conn:
                conn.close()

        for already_enrolled_user in self.enrolled_users:
            if already_enrolled_user.user_id == enrolled_user.user_id:
                self.enrolled_users.remove(already_enrolled_user)
                break

    def _renew_enrolled_users_cache(self):
        conn = None
        try:
            conn = Driver().open(self.mbcontext)
            saved_minutiaes = [(x.get_entry(0), x.get_entry(1)) for x in conn.select("select UserId, hex(FingerPrintMinutiae) from Users where FingerPrintMinutiae is not null")]
            self.enrolled_users = []
            for saved_minutiae in saved_minutiaes:
                fmd_data = bytes(bytearray.fromhex(saved_minutiae[1]))
                enrolled_user = User(saved_minutiae[0], len(fmd_data), fmd_data)
                self.enrolled_users.append(enrolled_user)
        finally:
            if conn:
                conn.close()

        return self.enrolled_users


class User:
    def __init__(self, user_id, fmd_buffer_size, fmd_buffer):
        self.user_id = user_id
        self.fmd_buffer_size = fmd_buffer_size
        self.fmd_buffer = fmd_buffer


class DigitalPersonaFingerPrintReader(FingerPrintReader):
    DPFPDD_SUCCESS = 0
    DPFPDD_E_MORE_DATA = 0x0d

    DPFJ_SUCCESS = 0
    DPFJ_E_MORE_DATA = 0xd
    DPFJ_E_INVALID_FID = 0x65

    MAX_DEVICE_NAME_LENGTH = 1024
    MAX_STR_LENGTH = 128

    DPFJ_FMD_ANSI_378_2004 = 0x001B0001

    DPFPDD_IMG_FMT_PIXEL_BUFFER = 0
    DPFPDD_IMG_PROC_DEFAULT = 0
    DPFPDD_QUALITY_GOOD = 0

    DPFJ_POSITION_RINDEX = 2

    DPFPDD_STATUS_READY = 0
    DPFPDD_STATUS_FAILURE = 3

    # Events
    EVT_PLACE_FINGER_LOGIN = "EVT_PLACE_FINGER_LOGIN"
    EVT_PLACE_FINGER_AUTHORIZATION = "EVT_PLACE_FINGER_AUTHORIZATION"
    EVT_PLACE_RIGHT_INDEX_FINGER = "EVT_PLACE_RIGHT_INDEX_FINGER"
    EVT_PLACE_LEFT_INDEX_FINGER = "EVT_PLACE_LEFT_INDEX_FINGER"
    EVT_BAD_FINGERPRINT_READ = "EVT_BAD_FINGERPRINT_READ"
    EVT_FINGER_ALREADY_REGISTERED = "EVT_FINGER_ALREADY_REGISTERED"
    EVT_ENROLLMENT_SUCCESS = "EVT_ENROLLMENT_SUCCESS"
    EVT_CLOSE_ALL_DIALOGS = "EVT_CLOSE_ALL_DIALOGS"
    EVT_FINGER_READ_SUCCESSFULLY_READ_AGAIN = "EVT_FINGER_READ_SUCCESSFULLY_READ_AGAIN"

    def __init__(self, enrolled_users_repository, call_back, threshold, finger_print_timeout, module_directory):
        # type: (EnrolledUsersRepository, FingerPrintReaderCallback) -> None
        super(DigitalPersonaFingerPrintReader, self).__init__()
        self.call_back = call_back
        self.enrolled_users_repository = enrolled_users_repository
        self.threshold = threshold
        self.finger_print_timeout = finger_print_timeout
        self.module_directory = module_directory

        bits = platform.architecture()[0]
        if bits == "32bit":
            dll_directory = os.path.join(os.path.abspath(self.module_directory), "x32")
        else:
            dll_directory = os.path.join(os.path.abspath(self.module_directory), "x64")

        self.dpfpdd = ctypes.WinDLL(os.path.join(dll_directory, "dpfpdd.dll"))
        self.dpfj = ctypes.WinDLL(os.path.join(dll_directory, "dpfj.dll"))

        self.dpfpdd_init = self.dpfpdd["dpfpdd_init"]

        self.dpfpdd_exit = self.dpfpdd["dpfpdd_exit"]

        self.dpfpdd_version = self.dpfpdd["dpfpdd_version"]
        self.dpfpdd_version.argtypes = [ctypes.POINTER(DfpddVersion)]
        self.dpfpdd_version.restype = ctypes.c_int

        self.dpfpdd_query_devices = self.dpfpdd["dpfpdd_query_devices"]
        self.dpfpdd_query_devices.argtypes = [ctypes.POINTER(ctypes.c_uint), ctypes.POINTER(DpfpddDevInfo)]
        self.dpfpdd_query_devices.restype = ctypes.c_int

        self.dpfpdd_open = self.dpfpdd["dpfpdd_open"]
        self.dpfpdd_open.argtypes = [ctypes.POINTER(ctypes.c_char), ctypes.POINTER(ctypes.c_void_p)]
        self.dpfpdd_open.restype = ctypes.c_int

        self.dpfpdd_close = self.dpfpdd["dpfpdd_close"]
        self.dpfpdd_close.argtypes = [ctypes.c_void_p]
        self.dpfpdd_close.restype = ctypes.c_int

        self.dpfpdd_get_device_status = self.dpfpdd["dpfpdd_get_device_status"]
        self.dpfpdd_get_device_status.argtypes = [ctypes.c_void_p, ctypes.POINTER(DpfpddDevStatus)]
        self.dpfpdd_get_device_status.restype = ctypes.c_int

        self.dpfj_start_enrollment = self.dpfj["dpfj_start_enrollment"]
        self.dpfj_start_enrollment.argtypes = [ctypes.c_void_p]
        self.dpfj_start_enrollment.restype = ctypes.c_int

        self.dpfpdd_capture = self.dpfpdd["dpfpdd_capture"]
        self.dpfpdd_capture.argtypes = [ctypes.c_void_p, ctypes.POINTER(DpfpddCaptureParam), ctypes.c_uint, ctypes.POINTER(DpfpddCaptureResult), ctypes.POINTER(ctypes.c_uint), ctypes.c_char_p]
        self.dpfpdd_capture.restype = ctypes.c_int

        self.dpfj_create_fmd_from_raw = self.dpfj["dpfj_create_fmd_from_raw"]
        self.dpfj_create_fmd_from_raw.argtypes = [ctypes.c_char_p, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint, ctypes.c_uint, ctypes.c_int, ctypes.c_uint, ctypes.c_int, ctypes.c_char_p, ctypes.POINTER(ctypes.c_uint)]
        self.dpfj_create_fmd_from_raw.restype = ctypes.c_int

        self.dpfj_add_to_enrollment = self.dpfj["dpfj_add_to_enrollment"]
        self.dpfj_add_to_enrollment.argtypes = [ctypes.c_int, ctypes.c_char_p, ctypes.c_uint, ctypes.c_uint]
        self.dpfj_add_to_enrollment.restype = ctypes.c_int

        self.dpfj_create_enrollment_fmd = self.dpfj["dpfj_create_enrollment_fmd"]
        self.dpfj_create_enrollment_fmd.argtypes = [ctypes.c_char_p, ctypes.POINTER(ctypes.c_uint)]
        self.dpfj_create_enrollment_fmd.restype = ctypes.c_int

        self.dpfj_finish_enrollment = self.dpfj["dpfj_finish_enrollment"]
        self.dpfj_finish_enrollment.argtypes = []
        self.dpfj_finish_enrollment.restype = ctypes.c_int

        self.dpfj_identify = self.dpfj["dpfj_identify"]
        self.dpfj_identify.argtypes = [ctypes.c_int, ctypes.c_char_p, ctypes.c_uint, ctypes.c_uint, ctypes.c_int, ctypes.c_uint, ctypes.POINTER(ctypes.c_char_p), ctypes.POINTER(ctypes.c_uint), ctypes.c_uint, ctypes.POINTER(ctypes.c_uint), ctypes.POINTER(DpfjCandidate)]
        self.dpfj_identify.restype = ctypes.c_int

    def enroll_user(self, user_id, pos_id, replace):
        p_device = self._init_device()

        try:
            ret = self.dpfj_start_enrollment(DigitalPersonaFingerPrintReader.DPFJ_FMD_ANSI_378_2004)
            if ret != DigitalPersonaFingerPrintReader.DPFJ_SUCCESS:
                raise DigitalPersonaFingerPrintException(ret, "Error calling dpfj_start_enrollment")

            try:
                self.call_back.finger_print_reader_callback(
                    DigitalPersonaFingerPrintReader.EVT_PLACE_RIGHT_INDEX_FINGER, pos_id)
                finish = False
                attempt = 1
                while not finish:
                    try:
                        fmd_buffer_size, fmd_buffer = self._capture_fmd(p_device)
                    except DigitalPersonaFingerPrintException as ex:
                        if ex.error_code & 0xFFFF == DigitalPersonaFingerPrintReader.DPFJ_E_INVALID_FID:
                            self.call_back.finger_print_reader_callback(
                                DigitalPersonaFingerPrintReader.EVT_BAD_FINGERPRINT_READ, pos_id)
                            continue
                        else:
                            raise ex

                    show_error = False
                    identified_user_id = self._identify_fmd(p_device, fmd_buffer_size, fmd_buffer)
                    if replace and identified_user_id is not None and identified_user_id != user_id:
                        show_error = True
                    elif not replace and identified_user_id is not None:
                        show_error = True

                    if show_error:
                        self.call_back.finger_print_reader_callback(
                            DigitalPersonaFingerPrintReader.EVT_FINGER_ALREADY_REGISTERED, pos_id)
                        raise DigitalPersonaFingerPrintException(-1, "Fingerprint already present in the database")

                    ret = self.dpfj_add_to_enrollment(DigitalPersonaFingerPrintReader.DPFJ_FMD_ANSI_378_2004, fmd_buffer, fmd_buffer_size, 0)
                    if ret & 0xFF == DigitalPersonaFingerPrintReader.DPFJ_SUCCESS:
                        finish = True
                    elif ret & 0xFF != DigitalPersonaFingerPrintReader.DPFJ_E_MORE_DATA:
                        raise DigitalPersonaFingerPrintException(ret, "Error calling dpfj_add_to_enrollment")
                    else:
                        self.call_back.finger_print_reader_callback(
                            DigitalPersonaFingerPrintReader.EVT_FINGER_READ_SUCCESSFULLY_READ_AGAIN, pos_id, attempt)

                    attempt += 1

                fmd_buffer_size = ctypes.c_uint(2500)
                fmd_buffer = ctypes.create_string_buffer(fmd_buffer_size.value)
                ret = self.dpfj_create_enrollment_fmd(fmd_buffer, ctypes.pointer(fmd_buffer_size))
                if ret != DigitalPersonaFingerPrintReader.DPFJ_SUCCESS:
                    raise DigitalPersonaFingerPrintException(ret, "Error calling dpfj_create_enrollment_fmd")
            finally:
                ret = self.dpfj_finish_enrollment()
                if ret != DigitalPersonaFingerPrintReader.DPFJ_SUCCESS:
                    raise DigitalPersonaFingerPrintException(ret, "Error calling dpfj_finish_enrollment")

            fmd_buffer_correct_size = bytearray()
            for i in range(0, fmd_buffer_size.value, 1):
                fmd_buffer_correct_size.append(fmd_buffer[i])
            fmd_buffer_correct_size = bytes(fmd_buffer_correct_size)

            enrolled_user = User(user_id, fmd_buffer_size.value, fmd_buffer_correct_size)

            if replace:
                self.enrolled_users_repository.remove_enrolled_user(enrolled_user, pos_id)

            self.enrolled_users_repository.save_enrolled_user(enrolled_user, pos_id)

            self.call_back.finger_print_reader_callback(DigitalPersonaFingerPrintReader.EVT_ENROLLMENT_SUCCESS, pos_id)

            return enrolled_user.fmd_buffer
        except Exception as ex:
            if ex is not DigitalPersonaFingerPrintException or ex.error_code not in [-1]:
                self.call_back.finger_print_reader_callback(DigitalPersonaFingerPrintReader.EVT_CLOSE_ALL_DIALOGS, pos_id)

            raise ex
        finally:
            self._finish_device(p_device)

    def authenticate(self, pos_id, authorization=False):
        p_device = self._init_device()
        try:
            if authorization:
                self.call_back.finger_print_reader_callback(DigitalPersonaFingerPrintReader.EVT_PLACE_FINGER_AUTHORIZATION, pos_id)
            else:
                self.call_back.finger_print_reader_callback(DigitalPersonaFingerPrintReader.EVT_PLACE_FINGER_LOGIN, pos_id)
            fmd_buffer_size, fmd_buffer = self._capture_fmd(p_device)

            user_id = self._identify_fmd(p_device, fmd_buffer_size, fmd_buffer)

            return user_id
        finally:
            self.call_back.finger_print_reader_callback(DigitalPersonaFingerPrintReader.EVT_CLOSE_ALL_DIALOGS, pos_id)
            self._finish_device(p_device)

    def check_device(self):
        p_device = self._init_device()

        try:
            device_status = DpfpddDevStatus()
            device_status.size = ctypes.sizeof(DpfpddDevStatus)
            device_status_pointer = ctypes.pointer(device_status)

            ret = self.dpfpdd_get_device_status(p_device.contents.value, device_status_pointer)
            if ret != DigitalPersonaFingerPrintReader.DPFPDD_SUCCESS:
                raise DigitalPersonaFingerPrintException(ret, "Error calling dpfpdd_get_device_status")

            if device_status.status != DigitalPersonaFingerPrintReader.DPFPDD_STATUS_READY:
                raise DigitalPersonaFingerPrintException(-3, "Device failure: {0}".format(device_status.status))

        finally:
            self._finish_device(p_device)

    def _init_device(self):
        ret = self.dpfpdd_init()
        if ret != DigitalPersonaFingerPrintReader.DPFPDD_SUCCESS:
            raise DigitalPersonaFingerPrintException(ret, "Error calling dpfpdd_init")

        version = DfpddVersion()
        version.size = ctypes.sizeof(DfpddVersion)

        cnt_devices = ctypes.c_uint(0)
        ret = self.dpfpdd_query_devices(ctypes.pointer(cnt_devices), None)
        if ret & 0xFF != DigitalPersonaFingerPrintReader.DPFPDD_E_MORE_DATA:
            raise DigitalPersonaFingerPrintException(ret, "Error calling dpfpdd_query_devices")
        else:
            devInfo = DpfpddDevInfo()
            devInfo.size = ctypes.sizeof(DpfpddDevInfo)
            ret = self.dpfpdd_query_devices(ctypes.pointer(cnt_devices), ctypes.pointer(devInfo))
            if ret == DigitalPersonaFingerPrintReader.DPFPDD_SUCCESS:
                # Device found
                p_device = ctypes.pointer(ctypes.c_void_p(0))
                ret = self.dpfpdd_open(devInfo.name, p_device)
                if ret != 0:
                    raise DigitalPersonaFingerPrintException(ret, "Error calling dpfpdd_open")
            else:
                raise Exception(ret, "Error calling dpfpdd_query_devices")

        return p_device

    def _finish_device(self, p_device):
        ret = self.dpfpdd_close(p_device.contents.value)
        if ret != DigitalPersonaFingerPrintReader.DPFPDD_SUCCESS:
            raise DigitalPersonaFingerPrintException(ret, "Error calling dpfpdd_close")

        ret = self.dpfpdd_exit()
        if ret != DigitalPersonaFingerPrintReader.DPFPDD_SUCCESS:
            raise DigitalPersonaFingerPrintException(ret, "Error calling dpfpdd_exit")

    def _capture_fmd(self, p_device):
        dev_capture_param = DpfpddCaptureParam()
        dev_capture_param.size = ctypes.sizeof(DpfpddCaptureParam)
        dev_capture_param.img_fmt = DigitalPersonaFingerPrintReader.DPFPDD_IMG_FMT_PIXEL_BUFFER
        dev_capture_param.image_proc = DigitalPersonaFingerPrintReader.DPFPDD_IMG_PROC_DEFAULT
        dev_capture_param.image_res = 500

        dev_capture_result = DpfpddCaptureResult()
        dev_capture_result.size = ctypes.sizeof(DpfpddCaptureResult)

        buffer_size = ctypes.c_uint(200000)
        image_buffer = ctypes.create_string_buffer(buffer_size.value)

        ret = self.dpfpdd_capture(p_device.contents.value,
                                  ctypes.pointer(dev_capture_param),
                                  self.finger_print_timeout,
                                  ctypes.pointer(dev_capture_result),
                                  ctypes.pointer(buffer_size),
                                  image_buffer)
        if ret != DigitalPersonaFingerPrintReader.DPFPDD_SUCCESS:
            raise DigitalPersonaFingerPrintException(ret, "Error calling dpfpdd_capture")

        if buffer_size.value == 200000:
            # Se o buffer nao foi modificado, saimos por timeout
            raise DigitalPersonaFingerPrintException(-2, "Timeout reading finferprint")

        fmd_buffer_size = ctypes.c_uint(2500)
        fmd_buffer = ctypes.create_string_buffer(fmd_buffer_size.value)
        ret = self.dpfj_create_fmd_from_raw(image_buffer,
                                            buffer_size,
                                            dev_capture_result.info.width,
                                            dev_capture_result.info.height,
                                            500,
                                            DigitalPersonaFingerPrintReader.DPFJ_POSITION_RINDEX,
                                            0,
                                            DigitalPersonaFingerPrintReader.DPFJ_FMD_ANSI_378_2004,
                                            fmd_buffer,
                                            fmd_buffer_size)
        if ret != DigitalPersonaFingerPrintReader.DPFJ_SUCCESS:
            raise DigitalPersonaFingerPrintException(ret, "Error calling dpfj_create_fmd_from_raw")

        return fmd_buffer_size, fmd_buffer

    def _identify_fmd(self, p_device, fmd_buffer_size, fmd_buffer):
        enrolled_users = self.enrolled_users_repository.load_all_enrolled_users()

        if len(enrolled_users) == 0:
            return None

        fmds_cnt = ctypes.c_uint(len(enrolled_users))
        fmds_data = (ctypes.c_char_p * fmds_cnt.value)()
        fmds_size = (ctypes.c_uint * fmds_cnt.value)()
        for i in range(0, fmds_cnt.value, 1):
            fmds_data[i] = ctypes.c_char_p(enrolled_users[i].fmd_buffer)
            fmds_size[i] = ctypes.c_uint(enrolled_users[i].fmd_buffer_size)

        candidate = DpfjCandidate()
        candidate.fmd_idx = 10
        candidate.size = ctypes.sizeof(DpfjCandidate) - 10
        candidates = ctypes.pointer(candidate)

        candidate_cnt = ctypes.pointer(ctypes.c_uint(1))
        ret = self.dpfj_identify(DigitalPersonaFingerPrintReader.DPFJ_FMD_ANSI_378_2004,
                                 fmd_buffer,
                                 fmd_buffer_size,
                                 ctypes.c_uint(0),
                                 DigitalPersonaFingerPrintReader.DPFJ_FMD_ANSI_378_2004,
                                 fmds_cnt,
                                 fmds_data,
                                 fmds_size,
                                 ctypes.c_uint(self.threshold),
                                 candidate_cnt,
                                 ctypes.cast(candidates, ctypes.POINTER(DpfjCandidate)))

        if ret != DigitalPersonaFingerPrintReader.DPFJ_SUCCESS:
            raise DigitalPersonaFingerPrintException(ret, "Error calling dpfj_identify")

        if candidate_cnt.contents.value == 0:
            return None

        return enrolled_users[candidates.contents.fmd_idx].user_id


class DigitalPersonaFingerPrintException(Exception):
    def __init__(self, error_code, message):
        super(DigitalPersonaFingerPrintException, self).__init__(message)
        self.error_code = error_code


class DfpddVerInfo(ctypes.Structure):
    _fields_ = [
        ('major', ctypes.c_int),
        ('minor', ctypes.c_int),
        ('maintenance', ctypes.c_int)
    ]


class DfpddVersion(ctypes.Structure):
    _fields_ = [
        ('size', ctypes.c_uint),
        ('lib_ver', DfpddVerInfo),
        ('api_ver', DfpddVerInfo)
    ]


class DpfpddHwDescr(ctypes.Structure):
    _fields_ = [
        ('vendor_name', ctypes.c_char * DigitalPersonaFingerPrintReader.MAX_STR_LENGTH),
        ('product_name', ctypes.c_char * DigitalPersonaFingerPrintReader.MAX_STR_LENGTH),
        ('serial_num', ctypes.c_char * DigitalPersonaFingerPrintReader.MAX_STR_LENGTH)
    ]


class DpfpddHwId(ctypes.Structure):
    _fields_ = [
        ('vendor_id', ctypes.c_ushort),
        ('product_id', ctypes.c_ushort)
    ]


class DpfpddHwVersion(ctypes.Structure):
    _fields_ = [
        ('hw_ver', DfpddVerInfo),
        ('fw_ver', DfpddVerInfo),
        ('bcd_rev', ctypes.c_ushort),
    ]


class DpfpddDevInfo(ctypes.Structure):
    _fields_ = [
        ('size', ctypes.c_uint),
        ('name', ctypes.c_char * DigitalPersonaFingerPrintReader.MAX_DEVICE_NAME_LENGTH),
        ('descr', DpfpddHwDescr),
        ('id', DpfpddHwId),
        ('ver', DpfpddHwVersion),
        ('modality', ctypes.c_uint),
        ('technology', ctypes.c_uint),
    ]


class DpfpddCaptureParam(ctypes.Structure):
    _fields_ = [
        ('size', ctypes.c_uint),
        ('image_fmt', ctypes.c_uint),
        ('image_proc', ctypes.c_uint),
        ('image_res', ctypes.c_uint)
    ]


class DpfpddImageInfo(ctypes.Structure):
    _fields_ = [
        ('size', ctypes.c_uint),
        ('width', ctypes.c_uint),
        ('height', ctypes.c_uint),
        ('res', ctypes.c_uint),
        ('bpp', ctypes.c_uint)
    ]


class DpfpddCaptureResult(ctypes.Structure):
    _fields_ = [
        ('size', ctypes.c_uint),
        ('success', ctypes.c_int),
        ('quality', ctypes.c_uint),
        ('score', ctypes.c_uint),
        ('info', DpfpddImageInfo)
    ]


class DpfjCandidate(ctypes.Structure):
    _fields_ = [
        ('size', ctypes.c_uint),
        ('fmd_idx', ctypes.c_uint),
        ('view_idx', ctypes.c_uint)
    ]


class DpfpddDevStatus(ctypes.Structure):
    _fields_ = [
        ('size', ctypes.c_uint),
        ('status', ctypes.c_uint),
        ('finger_detected', ctypes.c_uint),
        ('data', ctypes.c_byte)
    ]



from datetime import datetime, timedelta

from production.box._OrderChangerProductionBox import OrderChangerProductionBox


class RemoveReprocessedOrdersBox(OrderChangerProductionBox):
    def __init__(self, name, sons, reprocessed_time_limit, logger=None):
        super(RemoveReprocessedOrdersBox, self).__init__(name, sons, logger)
        self.reprocessed_time_limit = int(reprocessed_time_limit)
        self.logger = logger

    def change_order(self, order):
        if order.first_processing and not order.purged:
            order_creation_time = datetime.strptime(order.created_at[0:19], "%Y-%m-%dT%H:%M:%S")
            if (datetime.now() - order_creation_time) > timedelta(seconds=self.reprocessed_time_limit):
                self.logger.error("Order removed after reprocessing time limit. Order: \n{}".format(str(order)))
                return None

        return order



import json
from datetime import datetime

from typing import Dict

from interactor.event import EventInteractor
from model.log import LogEntry
from processor.event import EventProcessor


def _get_log_entry_from_pos_event(
        data,       # type: str
        event_type  # type: Dict
):
    json_data = json.loads(data)
    created_at = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")
    business_period = datetime.utcnow().strftime("%Y%m%d")
    return LogEntry(
            pos_id=json_data.get("pos_id", None),
            business_period=business_period,
            event_type=event_type,
            created_at_gmt=created_at,
            event_data=data
    )


class PosUpdaterEventProcessor(
        EventProcessor  # type: EventInteractor
):

    def parse_data(self, data, event_type):
        return _get_log_entry_from_pos_event(data, event_type)






import json
from logging import Logger

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union
)

from posot import OrderTaker
from sysactions import get_tender_types

from application.model.benefit import Action, RequiredItemOnSale, OrderBenefitInfo, ActionResult, Conditions
from application.model.conditions import Restrictions
from application.model.exceptions import (
    InvalidTenderIdException,
    InvalidAmountException,
    InvalidPercentageException,
    NotEffectiveDiscountException,
    InvalidQuantityException,
    AddItemsException
)
from application.model.order import Order, SaleLine
from application.model.transactions import AddBenefitToApplyUnitOfWork
from application.util import CompositionType


class DiscountService(object):
    def __init__(self, pos_ot, product_groups, logger):
        # type: (OrderTaker, Dict, Logger) -> None

        self.pos_ot = pos_ot
        self.product_groups = product_groups
        self.logger = logger

    def apply_discount_amount_on_total(self, action, order, restriction, available_discount_ids):
        # type: (Action, Order, Restrictions, set) -> ActionResult

        return self._apply_discount_on_total(action, order, restriction, available_discount_ids)

    def apply_discount_percentage_on_total(self, action, order, restriction, available_discount_ids):
        # type: (Action, Order, Restrictions, set) -> ActionResult

        return self._apply_discount_on_total(action, order, restriction, available_discount_ids, True)

    def apply_discount_amount_by_groups(self, action, order, restriction, discount_ids, apply_on_total):
        # type: (Action, Order, Restrictions, set, bool) -> ActionResult

        return self._apply_discount_by_items(action, order, restriction, discount_ids, False, apply_on_total, True)

    def apply_discount_percentage_by_groups(self, action, order, restriction, discount_ids, apply_on_total):
        # type: (Action, Order, Restrictions, set, bool) -> ActionResult

        return self._apply_discount_by_items(action, order, restriction, discount_ids, True, apply_on_total, True)

    def apply_discount_amount_by_items(self, action, order, restriction, discount_ids, apply_on_total):
        # type: (Action, Order, Restrictions, set, bool) -> ActionResult

        return self._apply_discount_by_items(action, order, restriction, discount_ids, False, apply_on_total)

    def apply_discount_percentage_by_items(self, action, order, restriction, discount_ids, apply_on_total):
        # type: (Action, Order, Restrictions, set, bool) -> ActionResult

        return self._apply_discount_by_items(action, order, restriction, discount_ids, True, apply_on_total)

    def apply_discount_amount_on_total_by_tender(self, action, order, restriction, by_items=False):
        # type: (Action, Order, Restrictions, bool) -> ActionResult

        return self.apply_discount_on_total_by_tender(action, order, restriction, by_items=by_items)

    def apply_discount_percentage_on_total_by_tender(self, action, order, restriction):
        # type: (Action, Order, Restrictions) -> ActionResult

        return self.apply_discount_on_total_by_tender(action, order, restriction, True)

    def apply_items_to_add(self, action, burnt_items, uow, restriction, on_total=False):
        # type: (Action, Dict, AddBenefitToApplyUnitOfWork, Restrictions, bool) -> ActionResult

        items_to_burn = {}
        benefit_applied = True

        for entry in action.value:
            part_code = entry.get("item")
            quantity = entry.get("quantity")
            free = bool(entry.get("free"))
            price = entry.get("price", None)
            tender_id = entry.get("tenderId", None)

            quantity = self._get_quantity_to_apply(burnt_items, quantity, uow)
            items_to_burn.update(burnt_items)
            for _ in range(0, quantity):
                action_burnt_items = uow.add_sale_line(
                        part_code, free, price, tender_id, items_to_burn, restriction, on_total, action.order
                )
                items_to_burn.update(action_burnt_items)
                benefit_applied = benefit_applied and not free

        return ActionResult([], items_to_burn, [], {}, benefit_applied, not on_total)

    def apply_tender_on_added_items(self, action, order, benefit_order_info):
        # type: (Action, Order, OrderBenefitInfo) -> ActionResult

        order_tender_ids = []

        if benefit_order_info is None:
            raise AddItemsException()

        for promotion_tender in benefit_order_info.promotion_tender:
            if action.order != promotion_tender.get("action_id"):
                continue

            tender_amount = promotion_tender.get("amount", 0)
            if tender_amount > 0:
                promotion_tender_id = promotion_tender.get("tender_id")
                if promotion_tender_id is None:
                    raise InvalidTenderIdException()

                tender_id = int(promotion_tender_id)
                pos_id = order.pos_id
                order_tender_id = self._do_order_tender(pos_id, tender_id, tender_amount)
                if order_tender_id:
                    order_tender_ids.append(order_tender_id)

        return ActionResult([], {}, order_tender_ids, {}, True, False)

    def _apply_discount_on_total(self, action, order, restriction, available_discount_ids, use_percentage=False):
        # type: (Action, Order, Restrictions, set, bool) -> ActionResult

        total_gross = self._get_total_gross_amount(order, restriction)
        amount = total_gross * (float(action.value) / 100) if use_percentage else float(action.value)
        amount = min(float(amount), self._get_order_max_discount_amount(order, restriction))
        if amount > 0:
            discount_id = available_discount_ids.pop()
            self.pos_ot.applyDiscount(discount_id, amount, forcebyitem=1)
            discount_by_line_number = self._split_discount_by_sale_line(order, restriction, str(discount_id), amount)

            return ActionResult([discount_id], dict(), [], discount_by_line_number, True, False)

        return ActionResult([], {}, [], {}, False, False)

    def _apply_discount_by_items(
        self,
        action,                 # type: Action
        order,                  # type: Order
        restriction,            # type: Restrictions
        usable_discount_ids,    # type: set
        use_percentage,         # type: bool
        apply_on_total,         # type: bool
        by_groups=False         # type: bool
    ):
        # type: (...) -> ActionResult

        applied_discounts = []
        discount_by_group = None
        group_quantity = None
        if by_groups:
            group_quantity = action.times_to_apply
            items_to_apply_discount, discount_by_group = self._get_items_by_groups_to_apply_discount(
                action,
                order.sale_lines,
                use_percentage
            )
        else:
            items_to_apply_discount = self._get_items_to_apply_discount(action, order.sale_lines)

        discount_amount_on_total = 0
        total_discount = 0
        item_values = {}
        discount_by_line_number = {}
        for sale_line in order.sale_lines:
            if not restriction.validate(sale_line) or (group_quantity is not None and group_quantity == 0):
                continue

            current_discount = items_to_apply_discount.get(int(sale_line.item_id))
            if current_discount is None:
                continue

            message = "Applying discount by items: {} / {} / {}"
            self.logger.info(message.format(current_discount, sale_line.key, sale_line.qty))
            remaining_sale_line_qty = sale_line.qty
            amount = 0

            while remaining_sale_line_qty > 0 and current_discount is not None:
                selected_discount = current_discount
                group = selected_discount["group"] if discount_by_group else ""
                if discount_by_group:
                    quantity_to_be_removed = min(remaining_sale_line_qty, discount_by_group[group]["quantity"])
                else:
                    quantity_to_be_removed = min(remaining_sale_line_qty, selected_discount["quantity"])

                if use_percentage:
                    total_value_to_remove = sale_line.unit_price * quantity_to_be_removed
                    if not apply_on_total:
                        max_value_to_apply = sale_line.current_unit_price * quantity_to_be_removed
                        total_value_to_remove = min(total_value_to_remove, max_value_to_apply)
                else:
                    total_value_to_remove = sale_line.current_unit_price * quantity_to_be_removed

                if not discount_by_group:
                    remaining_sale_line_qty -= quantity_to_be_removed
                    selected_discount["quantity"] -= quantity_to_be_removed
                    group_quantity = None if group_quantity is None else group_quantity - quantity_to_be_removed

                if use_percentage:
                    amount_to_add = total_value_to_remove * (selected_discount["amount"] / 100)
                else:
                    if discount_by_group:
                        amount_to_add = discount_by_group[group]["amount"]
                    else:
                        amount_to_add = quantity_to_be_removed * selected_discount["amount"]

                amount += min(amount_to_add, total_value_to_remove)
                if discount_by_group and amount > 0:
                    discount_by_group[group]["amount"] -= min(amount_to_add, total_value_to_remove)
                    discount_by_group[group]["quantity"] -= quantity_to_be_removed
                    remaining_sale_line_qty -= quantity_to_be_removed
                    selected_discount["quantity"] -= quantity_to_be_removed
                    group_quantity = None if group_quantity is None else group_quantity - quantity_to_be_removed

                if discount_by_group and (discount_by_group[group]["quantity"] == 0 or discount_by_group[group]["amount"] == 0):
                    break

                if group_quantity == 0 or selected_discount["quantity"] == 0 or selected_discount["amount"] == 0:
                    break

            amount = min(float(amount), self._get_order_max_discount_amount(order, restriction))
            message = "Applying discount by items: sale line amount - {} / {} / {}"
            self.logger.info(message.format(current_discount, sale_line.key, amount))
            if amount > 0:
                line_num_level = "{}.{}".format(
                    sale_line.line_number,
                    sale_line.level
                )
                if line_num_level in item_values.keys():
                    item_values[line_num_level] += amount
                else:
                    item_values[line_num_level] = amount

                if apply_on_total:
                    discount_amount_on_total += amount
                else:
                    discount_id = usable_discount_ids.pop()
                    self._apply_discount(amount, apply_on_total, discount_id, sale_line)
                    applied_discounts.append(discount_id)
                    discount_by_line_number[discount_id] = item_values
                    item_values = {}

            total_discount += amount

        if discount_amount_on_total > 0:
            max_discount_amount = self._get_order_max_discount_amount(order, restriction)
            if discount_amount_on_total > max_discount_amount:
                discount_amount_on_total = max_discount_amount

            discount_id = usable_discount_ids.pop()
            self._apply_discount(discount_amount_on_total, apply_on_total, discount_id, None)
            applied_discounts.append(discount_id)
            discount_by_line_number[discount_id] = item_values

        self.logger.info("Applying discount by items: final total discount - {}".format(total_discount))
        if total_discount > 0:
            return ActionResult(applied_discounts, {}, [], discount_by_line_number, True, False)
        else:
            self.logger.warn("Non effective discount. Total discount: {}".format(total_discount))
            raise NotEffectiveDiscountException()

    def apply_discount_on_total_by_tender(self, action, order, restriction, use_percentage=False, by_items=False):
        # type: (Action, Order, Restrictions, bool, bool) -> ActionResult

        order_tender_ids = []
        for entry in action.value:
            tender_id = int(entry.get("tenderId", -1))
            if not self._validate_tender(tender_id):
                raise InvalidTenderIdException()

            if by_items:
                self._apply_payments_by_items(entry, order, order_tender_ids, restriction, tender_id)
            else:
                self._apply_payment_on_order(entry, order, order_tender_ids, restriction, tender_id, use_percentage)

        return ActionResult([], {}, order_tender_ids, {}, True, False)

    def _apply_payments_by_items(self, entry, order, order_tender_ids, restriction, tender_id):
        quantity = self._get_quantity_to_apply_discount(order, restriction.benefit.conditions)
        if not quantity:
            return

        for sale_line in order.sale_lines:
            item_id = sale_line.item_id
            if restriction.validate(sale_line) and item_id == str(entry["item"]):
                available_quantity = min(sale_line.qty, quantity)
                amount = available_quantity * sale_line.item_price
                order_tender_id = self._do_order_tender(order.pos_id, tender_id, amount)
                if order_tender_id:
                    order_tender_ids.append(order_tender_id)
                    line_number = sale_line.line_number
                    self.pos_ot.setOrderCustomProperty(
                            "discountPayment", order_tender_id, "", line_number, "1", "0", item_id
                    )

                quantity -= available_quantity
                if quantity == 0:
                    break

    def _apply_payment_on_order(self, entry, order, order_tender_ids, restriction, tender_id, use_percentage):
        amount = self._get_amount(entry, order, restriction, use_percentage)
        if not amount:
            raise NotEffectiveDiscountException()
        order_tender_id = self._do_order_tender(order.pos_id, tender_id, amount)
        if order_tender_id:
            order_tender_ids.append(order_tender_id)

    def _apply_discount(self, amount, apply_on_total, discount_id, sale_line):
        if apply_on_total:
            self.pos_ot.applyDiscount(discount_id, amount, forcebyitem=1)
        else:
            line_number = sale_line.line_number
            context = sale_line.context
            level = sale_line.level
            item_id = sale_line.item_id
            self.pos_ot.applyDiscount(discount_id, amount, line_number, context, level, item_id)

    def _get_total_amount_with_restrictions(self, order, restriction):
        # type: (Order, Restrictions) -> float

        amount = 0
        discount_by_line_number = self._split_discount_by_sale_line(order, restriction, "999", 100)
        total_gross = order.total_gross
        for sale_line in order.sale_lines:
            line_num_level = "{}.{}".format(
                sale_line.line_number,
                sale_line.level
            )
            if line_num_level not in discount_by_line_number["999"]:
                continue

            tip = order.tip * (sale_line.item_price / total_gross)
            amount += (sale_line.item_price - self._get_applied_discount_values(order, sale_line) + tip)

        return amount

    @staticmethod
    def _get_applied_discount_values(order, sale_line):
        # type: (Order, SaleLine) -> float

        amount = 0
        if sale_line.level != "0":
            return 0

        for benefit in order.applied_benefits:
            if not benefit.discount_item_values:
                continue

            for value in benefit.discount_item_values.itervalues():
                line_num_level = "{}.{}".format(
                    sale_line.line_number,
                    sale_line.level
                )
                if line_num_level not in value:
                    continue

                amount += float(value[line_num_level])

        return amount

    def _get_total_gross_amount(self, order, restriction):
        # type: (Order, Restrictions) -> float

        if not restriction.exists():
            return float(order.total_gross + order.tip)

        return self._get_total_amount_with_restrictions(order, restriction)

    def _get_order_max_discount_amount(
        self,
        order,          # type: Order
        restriction     # type: Restrictions
    ):
        # type: (...) -> float

        order_xml = order.order_picture

        if order_xml is None:
            return 0

        due_total_amount = float(order_xml.get("dueAmount", 0))

        if not restriction.exists():
            due_amount = due_total_amount
        else:
            due_amount = self._get_total_amount_with_restrictions(order, restriction)

        if due_total_amount > due_amount + 0.01:
            return due_amount

        return due_amount - 0.01

    def _get_items_to_apply_discount(
            self,
            action,     # type: Action
            sale_lines  # type: List[SaleLine]
    ):
        # type: (...) -> Dict

        items_apply_discount = {}
        for entry in action.value:
            key = int(entry.get("item", -1))
            combos = [line for line in sale_lines if self._check_if_discount_apply_on_sons(str(key), line)]
            if len(combos) > 0:
                return self._get_sons_to_apply_discount(combos, sale_lines, entry, action)

            items_apply_discount = self._manager_item_to_apply_discount(items_apply_discount, key, entry, action)

        return items_apply_discount

    @staticmethod
    def _check_if_discount_apply_on_sons(
            item_id,    # type: str
            sale_line   # type: SaleLine
    ):
        # type: (...) -> bool

        item_price = sale_line.item_price
        return sale_line.item_id == item_id and sale_line.item_type == CompositionType.COMBO and item_price == 0

    def _get_sons_to_apply_discount(
            self,
            combos,         # type: List[SaleLine]
            sale_lines,     # type: List[SaleLine]
            entry,          # type: Dict
            action,         # type: Action,
            group="",       # type: Optional[str],
    ):
        # type: (...) -> Dict

        items_apply_discount = {}
        for combo in combos:
            context_combo = "{}.{}".format(combo.context, combo.item_id)
            for line in sale_lines:
                is_a_combo_son = line.line_number == combo.line_number and line.context.startswith(context_combo)
                if not is_a_combo_son or line.item_price == 0:
                    continue

                item_id = int(line.item_id)
                items = self._manager_item_to_apply_discount(items_apply_discount, item_id, entry, action, group)
                items_apply_discount = items

        return items_apply_discount

    def _get_items_by_groups_to_apply_discount(self, action, sale_lines, use_percentage=False):
        # type: (Action, List[SaleLine], bool) -> (dict, dict)

        items_groups_to_apply_discount = {}
        discount_by_group = None if use_percentage else {}

        for entry in action.value:
            group = entry.get("group", "")
            if group not in self.product_groups:
                continue

            if not use_percentage:
                quantity = int(entry.get("quantity", 0))
                quantity = min(action.times_to_apply, quantity) if action.times_to_apply else quantity
                discount_by_group[group] = dict(amount=float(entry.get("amount", 0)), quantity=quantity)

            for product_id in self.product_groups[group]:
                key = int(product_id)
                combos = [line for line in sale_lines if self._check_if_discount_apply_on_sons(str(key), line)]
                if len(combos) > 0:
                    items = self._get_sons_to_apply_discount(combos, sale_lines, entry, action, group)
                    items_groups_to_apply_discount.update(items)
                    continue
                items = self._manager_item_to_apply_discount(items_groups_to_apply_discount, key, entry, action, group)
                items_groups_to_apply_discount = items

        return items_groups_to_apply_discount, discount_by_group

    @staticmethod
    def _manager_item_to_apply_discount(
            items_to_apply_discount,  # type: Dict
            item_id,                  # type: int
            entry,                    # type: Dict,
            action,                   # type: Action,
            group="",                 # type: Optional[str],
    ):
        # type: (...) -> Dict

        amount = float(entry.get("amount", entry.get("percentage", 0)))
        quantity = int(entry.get("quantity", 0))
        quantity = min(action.times_to_apply, quantity) if action.times_to_apply else quantity

        if item_id not in items_to_apply_discount:
            items_to_apply_discount[item_id] = {"quantity": quantity, "amount": amount,  "group": group}

        return items_to_apply_discount

    def _get_amount(self, entry, order, restriction, use_percentage):
        order.get_table_info()
        total_gross = self._get_total_gross_amount(order, restriction)
        if use_percentage:
            amount = float(entry.get("percentage", 0))
            if amount == 0:
                raise InvalidPercentageException()

            amount = total_gross * (amount / 100)
        else:
            amount = float(entry.get("amount", 0))
            if amount == 0:
                raise InvalidAmountException()

        return min(amount, total_gross)

    @staticmethod
    def _get_quantity_to_apply_discount(order, conditions):
        # type: (Order, Conditions) -> int

        necessary_items = {}
        for required_item in conditions.required_items_on_sale:
            necessary_items[str(required_item.item_id)] = required_item.quantity

        order_necessary_items = {}
        for sale_line in order.sale_lines:
            if sale_line.item_id in necessary_items:
                if sale_line.item_id not in order_necessary_items:
                    order_necessary_items[sale_line.item_id] = 0

                order_necessary_items[sale_line.item_id] += sale_line.qty

        quantity = None
        for key, value in order_necessary_items.iteritems():
            x = value / necessary_items[key]
            quantity = min(quantity, x) if quantity else x

        return quantity

    def _get_quantity_to_apply(self, burnt_items, quantity, uow):
        # type: (Dict, int, AddBenefitToApplyUnitOfWork) -> int

        available_qty = []
        ignore = uow.benefit.conditions.ignore_burned_items
        for required in uow.benefit.conditions.required_items_on_sale:
            available_item_quantity = self._get_available_item_quantity(
                burnt_items={} if ignore else burnt_items,
                required=required,
                uow=uow
            )
            available_qty.append(available_item_quantity / required.quantity)

        for required in uow.benefit.conditions.required_groups_on_sale:
            available_item_quantity = self._get_available_item_quantity(
                burnt_items={} if ignore else burnt_items,
                required=required,
                uow=uow,
                is_group=True
            )
            available_qty.append(available_item_quantity / required.quantity)

        quantity *= min(available_qty) if available_qty else 1

        # TODO: Need to validate rule to handle cases where the item is sold by kg
        if quantity % 1 > 0:
            raise InvalidQuantityException()

        return int(quantity)

    def _get_available_item_quantity(
        self,
        burnt_items,    # type: Dict
        required,       # type: Any[Union[RequiredItemOnSale, str]]
        uow,            # type: AddBenefitToApplyUnitOfWork
        is_group=False  # type: Optional[bool]
    ):
        # type: (...) -> int

        available_quantity = 0
        for sale_line in uow.order.sale_lines:
            valid_context = is_group or required.context is None or sale_line.context == required.context
            item_in_group = is_group and int(sale_line.item_id) in self.product_groups.get(required.group, [])
            required_item_id = "" if is_group else str(required.item_id)
            if valid_context and (item_in_group or sale_line.item_id == required_item_id):
                available_quantity += sale_line.qty
                if sale_line.key in burnt_items:
                    available_quantity -= burnt_items[sale_line.key]

        return available_quantity

    def _do_order_tender(self, pos_id, tender_id, amount):
        # type: (int, int, float) -> Optional[int]

        order_tender = self.pos_ot.doTender(pos_id, tender_id, amount, donotclose=True)
        if order_tender:
            order_tender_id = int(order_tender.get("tenderId", 0))
            return None if order_tender_id == 0 else order_tender_id

    @staticmethod
    def _validate_tender(tender_id):
        # type: (int) -> bool

        all_tenders = get_tender_types()
        all_tenders = [key for key in all_tenders.keys()]
        return tender_id in all_tenders

    @staticmethod
    def _get_last_sale_line_discounts(order):
        sale_line_discounts = {}

        order_picture = order.order_picture
        benefit_list = order_picture.find(".//OrderProperty[@key=\"BENEFIT_LIST\"]")
        benefit_list = benefit_list.get("value") if benefit_list is not None else None

        benefits = json.loads(benefit_list) if benefit_list else {}

        for discount_id, line_detail in benefits.get("discountItemValues", {}).items():
            for line_num_level, amount in line_detail.items():
                if line_num_level not in sale_line_discounts.keys():
                    sale_line_discounts[line_num_level] = amount
                else:
                    sale_line_discounts[line_num_level] += amount

        return sale_line_discounts

    def _split_discount_by_sale_line(self, order, restriction, discount_id, discount_amount):
        # type: (Order, Restrictions, str, float) -> Dict

        last_sale_line_discounts = self._get_last_sale_line_discounts(order)

        discount_by_line_number = {discount_id: {}}
        valid_sale_lines = {}
        total_amount = 0.0
        for sale_line in order.sale_lines:
            if restriction.validate(sale_line):
                line_num_level = "{}.{}".format(
                    sale_line.line_number,
                    sale_line.level
                )
                price = sale_line.unit_price * sale_line.qty
                if last_sale_line_discounts.get(line_num_level, 0.0) >= price:
                    continue
                total_amount += price
                valid_sale_lines[line_num_level] = price

        if len(order.sale_lines) == len(valid_sale_lines.keys()):
            return {}
        else:
            for key, line_amount in valid_sale_lines.items():
                sale_line_amount = round((discount_amount * (line_amount / total_amount)), 2)
                discount_by_line_number[discount_id][key] = sale_line_amount

        return discount_by_line_number



# -*- coding: utf-8 -*-
import json

from sysactions import (
    action,
    get_model,
    get_custom,
    get_operator_session,
    show_messagebox
)
from bustoken import TK_RUPTURA_UPDATE_DAILY_DISHES
from msgbus import (
    TK_SYS_NAK,
    FM_PARAM
)

from .. import logger
from .. import mb_context
from model.customexception import UpdateRuptureException


@action
def save_daily_dishes(
    pos_id,                 # type: str
    removed_dishes_json,    # type: str
    added_dishes_json       # type: str
):
    # type: (...) -> None

    model = get_model(pos_id)

    try:
        removed_dishes = json.loads(removed_dishes_json)
        added_dishes = json.loads(added_dishes_json)
        dishes_changes_json = json.dumps(
            {
                'unset': removed_dishes,
                'set': added_dishes
            }
        )

        last_manager_authorization = get_custom(
            model=model,
            key='Last Manager ID'
        )
        operator = get_operator_session(model) or None
        user_id = last_manager_authorization or operator or ''
        formatted_user_id = '{}'.format(user_id)
        data = '\0'.join(map(str, [dishes_changes_json, formatted_user_id]))
        msg = mb_context.MB_EasySendMessage(
            dest_name="Ruptura",
            token=TK_RUPTURA_UPDATE_DAILY_DISHES,
            format=FM_PARAM,
            data=data
        )
        if msg.token == TK_SYS_NAK:
            raise UpdateRuptureException(msg.data)

        show_messagebox(
            posid=pos_id,
            message="$DAILY_DISHES_UPDATE_SUCCESS"
        )
    except Exception as e:
        logger.exception('[save_daily_dishes] Exception {}'.format(e))
        show_messagebox(
            posid=pos_id,
            message="$DAILY_DISHES_UPDATE_SUCCESS",
            buttons="$EXIT"
        )



class Account(object):
    def __init__(self, account_type, amount, tender_id, operation_date, glaccount):
        self.account_type = account_type
        self.amount = amount
        self.tender_id = tender_id
        self.operation_date = operation_date
        self.glaccount = glaccount



import unittest
import simplejson as json

dct1 = {
    'key1': 'value1'
}

dct2 = {
    'key2': 'value2',
    'd1': dct1
}

dct3 = {
    'key2': 'value2',
    'd1': json.dumps(dct1)
}

dct4 = {
    'key2': 'value2',
    'd1': json.RawJSON(json.dumps(dct1))
}


class TestRawJson(unittest.TestCase):

    def test_normal_str(self):
        self.assertNotEqual(json.dumps(dct2), json.dumps(dct3))

    def test_raw_json_str(self):
        self.assertEqual(json.dumps(dct2), json.dumps(dct4))
        self.assertEqual(dct2, json.loads(json.dumps(dct4)))

    def test_list(self):
        self.assertEqual(
            json.dumps([dct2]),
            json.dumps([json.RawJSON(json.dumps(dct2))]))
        self.assertEqual(
            [dct2],
            json.loads(json.dumps([json.RawJSON(json.dumps(dct2))])))

    def test_direct(self):
        self.assertEqual(
            json.dumps(dct2),
            json.dumps(json.RawJSON(json.dumps(dct2))))
        self.assertEqual(
            dct2,
            json.loads(json.dumps(json.RawJSON(json.dumps(dct2)))))



from application.model.exceptions import (
    ApplyingDiscountUnknownException,
    AskValueDiscountMaxValueReachedException,
    BenefitExpiredOrInactiveException,
    CustomBenefitAlreadyAppliedException,
    ErrorApplyingCoupon,
    ExcessiveSaleAmountException,
    InsufficientSaleAmountException,
    InvalidBrandException,
    InvalidMinSeatsException,
    InvalidSaleTypeException,
    InvalidSectorException,
    InvalidServicePeriodOrderException,
    InvalidServiceTypeException,
    MissingItemsException,
    PodTypeNotValidException,
    UserLevelNotAllowedException,
)

from abc import ABCMeta
from datetime import datetime, timedelta
from xml.etree import cElementTree as eTree

from application.model.benefit import Benefit, OrderBenefitInfo, Conditions, ActionTypes, Action
from application.model.order import Order, SaleTypes
from application.util import get_service_period
from iso8601 import iso8601
from typing import Dict, List


class BenefitValidator:
    __metaclass__ = ABCMeta

    @staticmethod
    def validate_benefits(uow, applied_benefits, burnt_items, product_groups):
        BenefitValidator.validate_benefit_static_info(applied_benefits, uow.benefit)
        BenefitValidator.validate_benefit_model_info(
            benefit=uow.benefit,
            model=uow.model,
            user_auth_level=uow.user_auth_level
        )
        BenefitValidator.validate_benefit_order_info(uow.benefit, uow.order, burnt_items, product_groups)
        BenefitValidator.validate_benefit_table_info(uow.benefit, uow.order)
        BenefitValidator.validate_benefit_actions(uow.benefit, uow.order, product_groups)

    @staticmethod
    def validate_benefit_static_info(applied_benefits, benefit):
        # type: (list, Benefit) -> None

        conditions = benefit.conditions
        BenefitValidator._is_benefit_valid_now(conditions)
        BenefitValidator._validate_unique_for_sale(applied_benefits, conditions)
        BenefitValidator._validate_single_use(benefit, applied_benefits)
        BenefitValidator._validate_ask_benefit(benefit)

    @staticmethod
    def validate_benefit_model_info(
        benefit,            # type: Benefit
        model,              # type: eTree.Element
        user_auth_level     # type: bool
    ):
        # type: (...) -> None

        conditions = benefit.conditions
        BenefitValidator._is_benefit_valid_for_business_period(conditions, model)
        BenefitValidator._validate_pod_type(conditions, model)
        BenefitValidator._validate_user_min_level(conditions, user_auth_level)

    @staticmethod
    def validate_benefit_order_info(benefit, order, burnt_items, product_groups):
        # type: (Benefit, Order, Dict, Dict) -> None

        conditions = benefit.conditions
        BenefitValidator._validate_sale_type(conditions, order)
        BenefitValidator._validate_sector(conditions, order)
        BenefitValidator._validate_min_sale_amount(conditions, order)
        BenefitValidator._validate_max_sale_amount(conditions, order)
        BenefitValidator._validate_required_items_on_sale(benefit, burnt_items, order)
        BenefitValidator._validate_required_groups_on_sale(benefit, burnt_items, order, product_groups)
        BenefitValidator._validate_amount_discount(benefit, order)
        BenefitValidator._validate_service_type(conditions, order)
        BenefitValidator._validate_service_period(conditions, order)
        BenefitValidator._validate_brand(conditions, order)

    @staticmethod
    def validate_benefit_table_info(benefit, order):
        # type: (Benefit, Order) -> None

        conditions = benefit.conditions
        BenefitValidator._validate_min_seats(conditions, order)

    @staticmethod
    def validate_benefit_actions(benefit, order, product_groups):
        for action in benefit.actions:
            if benefit.ask_value and action.value <= 0:
                raise ErrorApplyingCoupon("NDISCOUNT_ERROR_BENEFIT_MISSING_VALUE")
            if action.name == "itemsToAdd":
                continue

            if isinstance(action.value, list):
                for value in action.value:
                    if "item" in value:
                        required_item_id = value["item"]
                        for sale_line in order.sale_lines:
                            if int(sale_line.item_id) == int(required_item_id) and sale_line.qty > 0:
                                break
                        else:
                            raise MissingItemsException('NDISCOUNT_ERROR_REQUIRED_ITEMS_ON_SALE')
                    elif "group" in value:
                        required_group = value["group"]
                        if required_group not in product_groups:
                            raise MissingItemsException('NDISCOUNT_ERROR_REQUIRED_ITEMS_ON_SALE')

                        for sale_line in order.sale_lines:
                            if int(sale_line.item_id) in product_groups[required_group] and sale_line.qty > 0:
                                break
                        else:
                            raise MissingItemsException('NDISCOUNT_ERROR_REQUIRED_ITEMS_ON_SALE')

    @staticmethod
    def _validate_required_items_on_sale(benefit, burnt_items, order):
        if not BenefitValidator._enough_items_on_order(benefit, order, burnt_items):
            raise MissingItemsException('NDISCOUNT_ERROR_REQUIRED_ITEMS_ON_SALE')

    @staticmethod
    def _validate_required_groups_on_sale(benefit, burnt_items, order, product_groups):
        if not BenefitValidator._enough_items_groups_on_order(benefit, order, burnt_items, product_groups):
            raise MissingItemsException('NDISCOUNT_ERROR_REQUIRED_GROUPS_ON_SALE')

    @staticmethod
    def _validate_max_sale_amount(conditions, order):
        order_xml = order.order_picture
        due_amount = float(order_xml.get("dueAmount", 0))
        if conditions.max_sale_amount is not None and due_amount > conditions.max_sale_amount:
            raise ExcessiveSaleAmountException('NDISCOUNT_ERROR_MAX_SALE_AMOUNT')

    @staticmethod
    def _validate_min_sale_amount(conditions, order):
        order_xml = order.order_picture
        due_amount = float(order_xml.get("dueAmount", 0))
        if conditions.min_sale_amount is not None and due_amount < conditions.min_sale_amount:
            raise InsufficientSaleAmountException('NDISCOUNT_ERROR_MIN_SALE_AMOUNT')

    @staticmethod
    def _validate_unique_for_sale(applied_benefits, conditions):
        # type: (list, Conditions) -> None

        if len(applied_benefits) > 0:
            if conditions.unique_for_sale:
                raise CustomBenefitAlreadyAppliedException('NDISCOUNT_ERROR_UNIQUE_FOR_SALE')
            for applied_benefit in applied_benefits:
                benefit = Benefit(applied_benefit.get("benefit"))
                if benefit.conditions.unique_for_sale:
                    raise CustomBenefitAlreadyAppliedException('NDISCOUNT_ERROR_UNIQUE_FOR_SALE_HAS_EXISTS')

    @staticmethod
    def _is_benefit_valid_now(conditions):
        # type: (Conditions) -> None

        valid_hours = conditions.valid_hours
        now = datetime.now().isoformat().split("T")[1]
        if valid_hours.start_hour is not None and now < valid_hours.start_hour:
            raise BenefitExpiredOrInactiveException('NDISCOUNT_START_HOUR')
        if valid_hours.end_hour is not None and now > valid_hours.end_hour:
            raise BenefitExpiredOrInactiveException('NDISCOUNT_END_HOUR')

    @staticmethod
    def _is_benefit_valid_for_business_period(
        conditions,     # type: Conditions
        model,          # type: eTree.XML
    ):
        # type: (...) -> None

        period = model.find("PosState").get("period")
        datetime_period = iso8601.parse_date(period)

        if conditions.valid_days is not None and str(datetime_period.isoweekday() % 7 + 1) not in conditions.valid_days:
            raise BenefitExpiredOrInactiveException('NDISCOUNT_VALID_DAYS')

        if conditions.valid_start_period is not None and datetime_period < conditions.valid_start_period:
            raise BenefitExpiredOrInactiveException('NDISCOUNT_START_PERIOD')

        period_end = datetime_period + timedelta(days=1)
        if conditions.valid_end_period is not None and period_end > conditions.valid_end_period:
            raise BenefitExpiredOrInactiveException('NDISCOUNT_END_PERIOD')

    @staticmethod
    def _enough_items_on_order(benefit, order, burnt_items):
        # type: (Benefit, Order, dict) -> int

        sale_lines = order.sale_lines
        conditions = benefit.conditions

        required_items_on_sale = {}
        available_items_on_sale = {}

        for item in conditions.required_items_on_sale:
            key = "{};{}".format(item.context if item.context else 'any', str(item.item_id))
            required_items_on_sale[key] = item.quantity
            available_items_on_sale[key] = 0

        if not required_items_on_sale:
            return True

        for sale_line in sale_lines:
            context = sale_line.context
            available_quantity = sale_line.qty
            available_quantity -= 0 if conditions.ignore_burned_items else int(burnt_items.get(sale_line.key, 0))

            for item_index in required_items_on_sale.keys():
                required_item_context, required_item_id = item_index.split(';')
                if int(sale_line.item_id) != int(required_item_id):
                    continue

                if required_item_context == 'any' or required_item_context in context:
                    available_items_on_sale[item_index] += available_quantity

        times = min([available_items_on_sale[i] / required_items_on_sale[i] for i in required_items_on_sale.keys()])

        if times > 0:
            benefit.times_to_apply = times if conditions.burn_items else None
            return True

        return False

    @staticmethod
    def _enough_items_groups_on_order(benefit, order, burnt_items, product_groups):
        # type: (Benefit, Order, dict, dict) -> int

        sale_lines = order.sale_lines
        conditions = benefit.conditions

        required_groups_on_sale = {}
        available_groups_on_sale = {}

        for group in conditions.required_groups_on_sale:
            required_groups_on_sale[group.group] = group.quantity
            available_groups_on_sale[group.group] = 0

        if not required_groups_on_sale:
            return True

        for sale_line in sale_lines:
            available_quantity = sale_line.qty
            available_quantity -= 0 if conditions.ignore_burned_items else int(burnt_items.get(sale_line.key, 0))

            for group_index in required_groups_on_sale.keys():
                if int(sale_line.item_id) not in product_groups[group_index]:
                    continue

                available_groups_on_sale[group_index] += available_quantity

        times = min([available_groups_on_sale[i]/required_groups_on_sale[i] for i in required_groups_on_sale.keys()])

        if times > 0:
            benefit.times_to_apply = times if conditions.burn_items else None
            return True
        return False

    @staticmethod
    def validate_added_sale_line_benefit(applied_benefit, order):
        # type: (OrderBenefitInfo, Order) -> bool

        for sale_line in order.sale_lines:
            item_removed = int(sale_line.qty) == int(sale_line.level) == 0
            benefits_exists = int(sale_line.line_number) in applied_benefit.added_sale_lines
            if benefits_exists and item_removed:
                return False

        return True

    @staticmethod
    def _validate_pod_type(conditions, model):
        # type: (Conditions, eTree.XML) -> None

        if conditions.pod_type:
            try:
                pod_type = model.find(".//WorkingMode").get("podType")
                if pod_type is None:
                    raise PodTypeNotValidException('NDISCOUNT_ERROR_POD_TYPE')

                pod_types = [pod_type]
                if pod_type.upper() == "FL":
                    pod_types = ["FC", "TS"]

                if not any(x in pod_types for x in conditions.pod_type):
                    raise PodTypeNotValidException('NDISCOUNT_ERROR_POD_TYPE')
            except PodTypeNotValidException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException('NDISCOUNT_ERROR_UNKNOWN')

    @staticmethod
    def _validate_user_min_level(conditions, user_auth_level):
        # type: (Conditions, int) -> None

        if conditions.user_min_level is not None:
            try:
                if user_auth_level < conditions.user_min_level:
                    raise UserLevelNotAllowedException('NDISCOUNT_ERROR_USER_MIN_LEVEL')
            except UserLevelNotAllowedException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException('NDISCOUNT_ERROR_UNKNOWN')

    @staticmethod
    def _validate_service_type(conditions, order):
        # type: (Conditions, Order) -> None

        if conditions.service_type:
            try:
                service_type = "Table" if order.custom_properties.get("TABLE_ID") is not None else "Quick"
                if service_type not in conditions.service_type:
                    raise InvalidServiceTypeException("NDISCOUNT_ERROR_SERVICE_TYPE_" + service_type.upper())
            except InvalidServiceTypeException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException('NDISCOUNT_ERROR_UNKNOWN')

    @staticmethod
    def _validate_sale_type(conditions, order):
        # type: (Conditions, Order) -> None

        sale_types = conditions.sale_type
        if sale_types:
            try:
                order_sale_type = SaleTypes[order.sale_type].value if SaleTypes[order.sale_type] else 'ERROR'
                any_table_order_has_correct_sale_type = BenefitValidator._validate_table_sale_type(sale_types, order)

                if order_sale_type not in sale_types and not any_table_order_has_correct_sale_type:
                    raise InvalidSaleTypeException("NDISCOUNT_ERROR_SALE_TYPE")
            except InvalidSaleTypeException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException('NDISCOUNT_ERROR_UNKNOWN')

    @staticmethod
    def _validate_sector(conditions, order):
        # type: (Conditions, Order) -> None

        if conditions.sector:
            order.get_table_info()
            try:
                if order.sector is None:
                    raise InvalidSectorException("NDISCOUNT_ERROR_SECTOR_NOT_TABLE")

                if order.sector not in conditions.sector:
                    raise InvalidSectorException("NDISCOUNT_ERROR_SECTOR")
            except InvalidSectorException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException()

    @staticmethod
    def _validate_min_seats(conditions, order):
        # type: (Conditions, Order) -> None

        if conditions.min_seats:
            order.get_table_info()
            message = 'NDISCOUNT_ERROR_MIN_REQUIRED_SEATS|{0}'.format(conditions.min_seats)
            try:
                if conditions.min_seats > order.seats:
                    raise InvalidMinSeatsException(message)
            except InvalidMinSeatsException as _:
                raise
            except Exception as _:
                raise ApplyingDiscountUnknownException('NDISCOUNT_ERROR_UNKNOWN')

    @staticmethod
    def _validate_amount_discount(benefit_to_apply, order):
        # type: (Benefit, Order) -> None

        for action in benefit_to_apply.actions:
            if action.name == ActionTypes.discountAmountOnTotal.value:
                BenefitValidator.get_discount_on_total(action, order)
            if action.name == ActionTypes.discountPercentageOnTotal.value:
                BenefitValidator.get_discount_on_total(action, order, True)

    @staticmethod
    def get_discount_on_total(action, order, use_percentage=False):
        # type: (Action, Order, bool) -> None

        discount = order.total_gross * (float(action.value) / 100) if use_percentage else float(action.value)
        if discount >= order.total_gross:
            max_discount_amount = order.total_gross - 0.01
            message_value = 'NDISCOUNT_ERROR_MIN_REQUIRED_SALE_AMOUNT|{0:.2f}'.format(max_discount_amount)
            message_percentage = 'NDISCOUNT_ERROR_MAX_DISCOUNT_SALE_AMOUNT|{0:.0f}'.format(action.value)
            message = message_percentage if use_percentage else message_value
            raise InsufficientSaleAmountException(message)

    @staticmethod
    def _validate_single_use(benefit, applied_benefits):
        # type: (Benefit, list) -> None

        if benefit.conditions.single_use:
            for applied_benefit in applied_benefits:
                if applied_benefit["benefit"]["id"] == benefit.id:
                    raise CustomBenefitAlreadyAppliedException('NDISCOUNT_ERROR_SINGLE_USE')

    @staticmethod
    def _validate_ask_benefit(benefit):
        # type: (Benefit) -> None

        if benefit.ask_value:
            discount_max_value = benefit.conditions.discount_max_value
            if discount_max_value:
                total_discount_to_apply = sum([float(x.value) for x in benefit.actions])
                if total_discount_to_apply > discount_max_value:
                    raise AskValueDiscountMaxValueReachedException("NDISCOUNT_ASK_VALUE_MAX_VALUE_REACHED")

    @staticmethod
    def _validate_service_period(conditions, order):
        # type: (Conditions, Order) -> None

        if conditions.service_period and (order.order_state == 'IN_PROGRESS' or not order.table_id):
            current_service_period = get_service_period(order)
            current_service_period_id = current_service_period.get("servicePeriodId", None)
            if current_service_period_id not in conditions.service_period:
                raise InvalidServicePeriodOrderException('NDISCOUNT_INVALID_SERVICE_PERIOD_ORDER')

    @staticmethod
    def _validate_table_sale_type(conditions_sale_types, order):
        # type: (List[str], Order) -> bool

        if order.table_id is None:
            return False

        for sale_line in order.sale_lines:
            if sale_line.sale_type is None:
                continue
            sale_type = SaleTypes[sale_line.sale_type].value if SaleTypes[sale_line.sale_type] else 'ERROR'
            if sale_type in conditions_sale_types:
                return True
        return False

    @staticmethod
    def _validate_brand(conditions, order):
        # type: (Conditions, Order) -> None

        if conditions.brands and order.brand not in conditions.brands:
            raise InvalidBrandException("NDISCOUNT_ERROR_BRAND")



from logging import Logger

from model.error_type import PosValidationErrorsType
from interactor.route.order.create.dto import (
    OrderCreateRequest,
    OrderCreateResponse
)
from model.config import Config
from model.exception import ValidationError
from model.repository import PosRepositoryModel
from validation_util import check_business_period, validate_api_key


class OrderCreateInteractor(object):
    def __init__(self, logger, config, pos_repository):
        # type: (Logger, Config, PosRepositoryModel) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository

    def execute(self, request):
        # type: (OrderCreateRequest) -> OrderCreateResponse

        pos_id = validate_api_key(request, self.config.api_key_mapping)
        model = self.pos_repository.get_model(pos_id)
        pos_ot = self.pos_repository.get_pos_ot(model)
        check_business_period(self.pos_repository, pos_id, self.config, pos_ot)

        if self.pos_repository.has_current_order(model):
            raise ValidationError(PosValidationErrorsType.OrderAlreadyInProgress)

        order_id = self.pos_repository.create_order(pos_id, pos_ot, request.price_list, request.sale_type)
        self.logger.info("OrderId: {}".format(order_id))

        return OrderCreateResponse(order_id=order_id)



import datetime
from xml.etree import cElementTree as eTree

from model.config import Config, PosMap
from model.error_type import AuthErrorsType, PosValidationErrorsType, OperationErrorsType
from model.exception import AuthValidationErrors, ValidationError, OperationErrors
from model.repository import PosRepositoryModel
from typing import List
from posot import OrderTaker

global api_key_reference


def check_if_business_day_is_opened(repository, model):
    # type: (PosRepositoryModel, eTree) -> None

    is_day_opened = repository.is_day_opened(model)
    if not is_day_opened:
        raise ValidationError(PosValidationErrorsType.BusinessDayNotOpened)


def check_if_operator_is_logged(repository, model):
    # type: (PosRepositoryModel, eTree) -> None

    is_operator_logged = repository.is_operator_logged(model)
    if not is_operator_logged:
        raise ValidationError(PosValidationErrorsType.OperatorIsNotLogged)


def check_if_order_is_in_progress(repository, model):
    # type: (PosRepositoryModel, eTree) -> None

    is_order_in_progress = repository.is_order_in_progress(model)
    if not is_order_in_progress:
        raise ValidationError(PosValidationErrorsType.OrderIsNotInProgress)


def check_if_order_is_totaled(repository, model):
    # type: (PosRepositoryModel, eTree) -> None

    is_order_totaled = repository.is_order_totaled(model)
    if not is_order_totaled:
        raise ValidationError(PosValidationErrorsType.OrderIsNotTotaled)


def check_if_order_is_in_progress_or_totaled(
        repository,     # type: PosRepositoryModel
        model           # type: eTree
):
    # type: (...) -> None

    is_order_in_progress = repository.is_order_in_progress(model)
    is_order_totaled = repository.is_order_totaled(model)

    if not (is_order_in_progress or is_order_totaled):
        raise ValidationError(PosValidationErrorsType.OrderIsNotInProgressOrTotaled)


def check_if_pos_id_is_valid(repository, pos_id):
    # type: (PosRepositoryModel, int) -> None

    pos_list = repository.get_pos_list()
    if not int(pos_id) in pos_list:
        raise ValidationError(PosValidationErrorsType.PosIdNotFound)


def check_if_payment_amount_is_valid(payment_value, model):
    # type: (float, eTree) -> None

    due_amount = float(model.find("CurrentOrder/Order").get("dueAmount"))
    if (payment_value <= 0) or (payment_value > due_amount):
        raise ValidationError(PosValidationErrorsType.PaymentAmountInvalid)


def check_if_payment_type_is_valid(allowed_methods, payment_type):
    # type: (list, int) -> None

    if payment_type not in allowed_methods:
        raise ValidationError(PosValidationErrorsType.PaymentTypeInvalid)


def check_if_due_amount_is_zero(model):
    # type: (eTree) -> None

    due_amount = float(model.find("CurrentOrder/Order").get("dueAmount"))
    if due_amount != 0:
        raise ValidationError(PosValidationErrorsType.DueAmountInvalid)


def check_if_operator_is_loggedin_and_logout(repository, operator, model, pos_id, pos_info, pos_ot):
    if operator is not None and operator.get("state") == "LOGGEDIN":
        if repository.has_current_order(model):
            repository.void_order(pos_id, pos_ot)

        operator_id = operator.get("id")
        if int(operator_id) != pos_info.user_id:
            logouted = repository.logout_delivery_user(pos_id, operator_id)
            if not logouted:
                raise OperationErrors(OperationErrorsType.OperatorLogoutError)
            operator = None

    return operator


def check_current_date_and_model_state_and_fix(repository, operator, model, pos_id, current_date):
    if model.find("PosState").get("period") != current_date or model.find("PosState").get("state") != "OPENED":

        if operator is not None and operator.get("state") == "LOGGEDIN":
            operator_id = operator.get("id")
            logouted = repository.logout_delivery_user(pos_id, operator_id)
            if not logouted:
                raise OperationErrors(OperationErrorsType.OperatorLogoutError)
            operator = None

        if model.find("PosState").get("state") == "OPENED" or model.find("PosState").get("state") == "BLOCKED":
            closed = repository.close_business_period(pos_id)
            if not closed:
                raise OperationErrors(OperationErrorsType.BusynessDayCloseError)

        opened = repository.open_business_period(pos_id)
        if not opened:
            raise OperationErrors(OperationErrorsType.BusynessDayOpenError)

    return operator


def check_if_operator_is_loggouted_and_login(repository, operator, pos_id, pos_info):
    if operator is None or operator.get("state") != "LOGGEDIN":
        authenticate_msg = repository.authenticate_user(str(pos_info.user_id), str(pos_info.user_password))

        if not authenticate_msg:
            raise OperationErrors(OperationErrorsType.OperatorLoginError)

        level, username, long_user_name = authenticate_msg.split('\x00')[:3]

        if operator is None or (operator.get("state") != "OPENED" and operator.get("state") != "PAUSED"):
            opened = repository.open_user(pos_id, str(pos_info.user_id), long_user_name)
            if not opened:
                raise OperationErrors(OperationErrorsType.BusynessDayOpenError)

        logged = repository.login_user(pos_id, str(pos_info.user_id), long_user_name)
        if not logged:
            raise OperationErrors(OperationErrorsType.OperatorLoginError)


def check_business_period(repository, pos_id, config, pos_ot):
    # type: (PosRepositoryModel, int, Config, OrderTaker) -> None

    model = repository.get_model(pos_id)
    current_date = datetime.datetime.now().date().strftime("%Y%m%d")
    operator = model.find("Operator")
    pos_info = next((x for x in config.api_key_mapping if x.pos_id == int(pos_id)), None)  # type: PosMap

    operator = check_if_operator_is_loggedin_and_logout(repository, operator, model, pos_id, pos_info, pos_ot)
    model = repository.get_model(pos_id)

    operator = check_current_date_and_model_state_and_fix(repository, operator, model, pos_id, current_date)

    check_if_operator_is_loggouted_and_login(repository, operator, pos_id, pos_info)


def validate_api_key(request, secure_api_key_list):
    api_key_accepted = False
    pos_id_accepted = False
    if hasattr(request, "pos_id"):
        received_pos_id = int(request.pos_id)
    else:
        received_pos_id = None
    received_api_key = request.system_api_key
    real_pos_id = ""

    for key in secure_api_key_list:
        if received_pos_id is not None:
            if key.api_key == received_api_key:
                api_key_accepted = True
                if received_pos_id == key.abstract_pos_id:
                    pos_id_accepted = True
                    real_pos_id = int(key.pos_id)
        elif key.api_key == received_api_key:
            api_key_accepted = True

    if not api_key_accepted:
        raise AuthValidationErrors(AuthErrorsType.InvalidApiKey)

    if received_pos_id and not pos_id_accepted:
        raise ValidationError(PosValidationErrorsType.PosIdNotFound)

    return str(real_pos_id)


def get_pos_id_from_pos_mapper(abstracted_pos_id, api_key_mapping):
    # type: (str, List[PosMap]) -> str

    for pos_map in api_key_mapping:
        if pos_map.abstract_pos_id == abstracted_pos_id:
            return pos_map.pos_id


def get_abstracted_pos_id_from_pos_mapper(pos_id, api_key_mapping):
    # type: (str, List[PosMap]) -> str

    for pos_map in api_key_mapping:
        if pos_map.pos_id == pos_id:
            return pos_map.abstract_pos_id



# -*- coding: utf-8 -*-
import datetime
import time

from xml.etree import cElementTree as eTree

import msgbus
import sysactions
import persistence
from actions.models import AuthorizationResponse

from actions.models.mw_sys_errors import MwSysErrors
from actions.util import get_operator_info, get_user_level, get_authorization
from typing import Dict

from .. import get_pos_service, mb_context, logger


def login_user(pos_id, user_id):
    try:
        long_name, level, user_id = get_user_info(user_id)

        msg = sysactions.send_message(
            get_pos_service(pos_id),
            msgbus.TK_POS_USERLOGIN,
            msgbus.FM_PARAM, "{}\0{}\0{}".format(pos_id, user_id, long_name),
            timeout=60 * 1000000
        )
        if msg.token != msgbus.TK_SYS_ACK:
            model = sysactions.get_model(pos_id)
            message = sysactions.translate_message(model, "ERROR") + ": "
            error = MwSysErrors.get_name(msg.data.replace("\0", ""))
            _handle_current_order_error(error, model, pos_id)
            message += sysactions.translate_message(model, error)
            sysactions.show_messagebox(pos_id, message, "$INFORMATION")
            logger.error("Invalid response from PosCtrl while login user: {} - {}".format(msg.token, msg.data))
            return False
    except (Exception, BaseException):
        logger.error("Exception while login user")
        sysactions.show_messagebox(pos_id, "$CANNOT_LOGIN_OPERATOR")
        return False
    return True


def _handle_current_order_error(error, model, pos_id):
    # type: (str, eTree.XML, int) -> None

    try:
        if error == "SE_ORDEROPENED":
            pos_ot = sysactions.get_posot(model)
            pos_ot.voidOrder(pos_id)
    except Exception as _:
        logger.exception("Error voiding order on user login")


def open_user(pos_id, user_id):
    msg = None
    try:
        long_name, level, user_id = get_user_info(user_id)

        msg = sysactions.send_message(
            get_pos_service(pos_id),
            msgbus.TK_POS_USEROPEN,
            msgbus.FM_PARAM,
            "%s\0%s\0%s\0%s\0%s" % (pos_id, user_id, 0.0, long_name, level),
            timeout=60 * 1000000
        )
        if msg.token != msgbus.TK_SYS_ACK:
            raise Exception()

    except (Exception, BaseException):
        logger.exception("Error opening user {} on POS {}".format(user_id, pos_id))
        message = "$CANNOT_OPEN_OPERATOR"
        if msg is not None and \
                msg.data and \
                len(msg.data.split("\0")) > 1 and \
                msg.data.split("\0")[0] in MwSysErrors.list_values():
            model = sysactions.get_model(pos_id)
            message = sysactions.translate_message(model, MwSysErrors.get_name(msg.data.split("\0")[0]))
        sysactions.show_messagebox(pos_id, message)
        return False
    return True


def close_user(pos_id, user_id):
    try:
        msg = sysactions.send_message(
            get_pos_service(pos_id),
            msgbus.TK_POS_USERLOGOUT,
            msgbus.FM_PARAM,
            "%s\0%s" % (pos_id, user_id),
            timeout=60 * 1000000
        )
        if msg.token != msgbus.TK_SYS_ACK:
            if msg.data.startswith("5007"):
                return True
            logger.exception("Error closing user: Token: {} / Data: {}".format(msg.token, msg.data))
            sysactions.show_messagebox(pos_id, "$CANNOT_CLOSE_OPERATOR")
            return False
    except (Exception, BaseException):
        logger.exception("Error closing user {} on POS {}".format(user_id, pos_id))
        sysactions.show_messagebox(pos_id, "$CANNOT_CLOSE_OPERATOR")
        return False

    return True


def check_current_day(pos_id):
    model = sysactions.get_model(pos_id)
    if sysactions.is_day_blocked(model):
        sysactions.show_messagebox(pos_id, "$POS_IS_BLOCKED_BY_TIME", "$INFORMATION")
        return False

    if not sysactions.is_day_opened(model):
        sysactions.show_messagebox(pos_id, "$NEED_TO_OPEN_BDAY_FIRST", "$INFORMATION")
        return False

    return True


def get_user_info(user_id):
    msg = sysactions.send_message("UserControl", msgbus.TK_USERCTRL_GETINFO, msgbus.FM_PARAM, user_id)
    if msg.token != msgbus.TK_SYS_ACK:
        raise Exception("Error getting user information {}-{}".format(msg.token, msg.data))

    user_info = eTree.XML(msg.data)
    long_name = user_info.find("user").get("LongName")
    level = user_info.find("user").get("Level")
    user_id = user_info.find("user").get("UserId")

    return long_name, level, user_id


def print_open_day_report(pos_id, initial_float):
    model = sysactions.get_model(pos_id)
    operator = sysactions.get_current_operator(model)
    operator_name = (operator.get("name")).encode('utf-8')
    operator_id = operator.get("id")
    period = sysactions.get_business_period(model)
    authorized_by = sysactions.get_custom(model, 'Last Manager ID', operator_id)

    sysactions.print_report(
            pos_id,
            model,
            True,
            "login_operator_report",
            pos_id,
            operator_id,
            operator_name,
            initial_float,
            period,
            authorized_by
    )


def get_pod_type_and_pos_function(pos_id):
    pod_type = sysactions.get_cfg(pos_id).key_value("pod") or 'FL'
    pos_function_config = sysactions.get_cfg(pos_id).key_value("posFunction") or 'FL'
    pos_function_list = pos_function_config.split("|")
    pos_function = pos_function_list[0]
    if len(pos_function_list) > 1:
        pos_function_index = sysactions.show_listbox(pos_id, pos_function_list, message="$SELECT_A_POS_FUNCTION")

        if pos_function_index is None:
            return None, None

        pos_function = pos_function_list[pos_function_index]

    return pod_type, pos_function


def change_pod_type_and_pos_function(pos_id, pod_type, pos_function):
    if pos_function == "DS":
        sysactions.check_main_screen(pos_id)
    sysactions.send_message(
        get_pos_service(pos_id),
        msgbus.TK_POS_SETPOD,
        msgbus.FM_PARAM, "{}\0{}\0".format(pos_id, pod_type),
        timeout=60 * 1000000)
    sysactions.send_message(
        get_pos_service(pos_id),
        msgbus.TK_POS_SETFUNCTION,
        msgbus.FM_PARAM, "{}\0{}\0".format(pos_id, pos_function),
        timeout=60 * 1000000)


def get_initial_float(pos_id, pod_type, pos_function):
    if pod_type != "OT" and pos_function != "OT":
        index = None
        list_min_values_drawer = sysactions.get_storewide_config("Store.MinValuesDrawer", defval=None)
        if list_min_values_drawer:
            list_limits = [("R$%.2f" % float(x)).replace(".", ",") for x in list_min_values_drawer.split(";")]
            if len(list_limits) > 1:
                index = sysactions.show_listbox(pos_id, list_limits, message="$ENTER_THE_INITIAL_FLOAT_AMOUNT",
                                                title="$OPERATOR_LOGIN", buttons="$OK|$CANCEL", icon="info",
                                                timeout=720000)
            elif len(list_limits) == 1:
                index = 0

            if index is None:
                raise sysactions.StopAction()

            initial_float = list_min_values_drawer.split(";")[index]
        else:
            initial_float = sysactions.show_keyboard(pos_id, message="$ENTER_THE_INITIAL_FLOAT_AMOUNT", title="$OPERATOR_OPENING",
                                                     mask="CURRENCY", numpad=True, timeout=720000)
            if initial_float is None:
                raise sysactions.StopAction()

        initial_float = round(float(initial_float or 0.0), 2)
        return float(initial_float)
    else:
        return 0.0


def save_initial_float(initial_float, pos_id, user_id):
    try:
        wait_user_logged_in(pos_id, user_id)
        model = sysactions.get_model(pos_id)
        _, operator_id, session = get_operator_info(model, user_id)
        sql = "insert into Transfer (Period, PosId, SessionId, Type, Description, Amount, TenderId) " \
              "values ({}, {}, '{}', 1, 'INITIAL_AMOUNT', {}, 0);"\
            .format(datetime.datetime.now().strftime('%Y%m%d'), pos_id, session, initial_float)
        persistence.Driver().open(mb_context, pos_id).query(sql)
    except Exception as ex:
        logger.exception('Fail set initial float to user: {} - ERROR: {}', user_id, ex)


def wait_user_logged_in(pos_id, user_id):
    # type: (str, str) -> None

    time_limit = (time.time() + 1.5)
    while (time.time() < time_limit) and (not has_session(sysactions.get_model(pos_id), user_id)):
        time.sleep(0.1)


def has_session(model, user_id):
    # type: (eTree, str) -> bool

    _, _, session = get_operator_info(model, user_id)
    return session not in ['', None]


def user_inserted_initial_amount(pos_id, session):
    try:
        if session in ['', None]:
            return False

        sql = "select count(*) from Transfer " \
              "where SessionId = '{}' and Description = 'INITIAL_AMOUNT'".\
            format(session)
        conn = persistence.Driver().open(mb_context, pos_id)
        row = conn.select(sql).get_row(0)
        return int(row.get_entry(0)) > 0
    except Exception as _:
        return False


def get_authorization_if_necessary(
        model,      # type: eTree.XML
        pos_id,     # type: str
        user_info,  # type: Dict
        min_level   # type: int
):
    # type: (...) -> AuthorizationResponse

    user_name = user_info["UserName"]
    user_level = int(user_info["Level"] or 0)
    if user_level < min_level:
        response = get_authorization(pos_id, min_level=min_level, model=model)
        if not response[0]:
            return AuthorizationResponse(False, "", "")

        user_name = response[1]

    return AuthorizationResponse(True, user_name, get_user_level(user_name))


def save_last_authorization(
        pos_id,                 # type: str
        authorized_by,          # type: str
        authorized_by_level     # type: str
):
    # type: (...) -> None

    sysactions.set_custom(pos_id, 'Last Manager ID', authorized_by)
    sysactions.set_custom(pos_id, 'Authorization Level', authorized_by_level)


def validate_user_level(
        pos_id,     # type: str
        user_info,  # type: Dict
):
    # type: (...) -> None

    if int(user_info.get("Level", -1)) < 0:
        logger.error("Access Denied for user: {}".format(user_info["UserName"]))
        sysactions.show_messagebox(pos_id, "$ACCESS_DENIED", "$INFORMATION")
        raise sysactions.StopAction()
